import numpy as np


class RectManager:
    """
    Backend management of a rect. Allows converting between rects and extents, also works with fractional inputs.
    """

    def __init__(self, x: float, y: float, w: float, h: float, canvas_rect: tuple):
        # initialize rect state arrays
        # used to store internal state of the rect in both fractional screen space and absolute screen space
        # the purpose of storing the fractional rect is that it remains constant when the canvas resizes
        self._rect_frac = np.zeros(4, dtype=np.float64)
        self._rect_screen_space = np.zeros(4, dtype=np.float64)
        self._canvas_rect = np.asarray(canvas_rect)

        self._set((x, y, w, h))

    def _set(self, rect):
        """
        Using the passed rect which is either absolute screen space or fractional,
        set the internal fractional and absolute screen space rects
        """
        rect = np.asarray(rect)
        for val, name in zip(rect, ["x-position", "y-position", "width", "height"]):
            if val < 0:
                raise ValueError(
                    f"Invalid rect value < 0: {rect}\n All values must be non-negative."
                )

        if (rect[2:] <= 1).all():  # fractional bbox
            self._set_from_fract(rect)

        elif (rect[2:] > 1).all():  # bbox in already in screen coords coordinates
            self._set_from_screen_space(rect)

        else:
            raise ValueError(f"Invalid rect: {rect}")

    def _set_from_fract(self, rect):
        """set rect from fractional representation"""
        _, _, cw, ch = self._canvas_rect
        mult = np.array([cw, ch, cw, ch])

        # check that widths, heights are valid:
        if rect[0] + rect[2] > 1:
            raise ValueError(
                f"invalid fractional rect: {rect}\n x + width > 1: {rect[0]} + {rect[2]} > 1"
            )
        if rect[1] + rect[3] > 1:
            raise ValueError(
                f"invalid fractional rect: {rect}\n y + height > 1: {rect[1]} + {rect[3]} > 1"
            )

        # assign values to the arrays, don't just change the reference
        self._rect_frac[:] = rect
        self._rect_screen_space[:] = self._rect_frac * mult

    def _set_from_screen_space(self, rect):
        """set rect from screen space representation"""
        _, _, cw, ch = self._canvas_rect
        mult = np.array([cw, ch, cw, ch])
        # for screen coords allow (x, y) = 1 or 0, but w, h must be > 1
        # check that widths, heights are valid
        if rect[0] + rect[2] > cw:
            raise ValueError(
                f"invalid rect: {rect}\n x + width > canvas width: {rect[0]} + {rect[2]} > {cw}"
            )
        if rect[1] + rect[3] > ch:
            raise ValueError(
                f"invalid rect: {rect}\n y + height > canvas height: {rect[1]} + {rect[3]} >{ch}"
            )

        self._rect_frac[:] = rect / mult
        self._rect_screen_space[:] = rect

    @property
    def x(self) -> np.float64:
        """x position"""
        return self._rect_screen_space[0]

    @property
    def y(self) -> np.float64:
        """y position"""
        return self._rect_screen_space[1]

    @property
    def w(self) -> np.float64:
        """width"""
        return self._rect_screen_space[2]

    @property
    def h(self) -> np.float64:
        """height"""
        return self._rect_screen_space[3]

    @property
    def rect(self) -> np.ndarray:
        """rect, (x, y, w, h)"""
        return self._rect_screen_space

    @rect.setter
    def rect(self, rect: np.ndarray | tuple):
        self._set(rect)

    def canvas_resized(self, canvas_rect: tuple):
        # called by Frame when canvas is resized
        self._canvas_rect[:] = canvas_rect
        # set new rect using existing rect_frac since this remains constant regardless of resize
        self._set(self._rect_frac)

    @property
    def x0(self) -> np.float64:
        """x0 position"""
        return self.x

    @property
    def x1(self) -> np.float64:
        """x1 position"""
        return self.x + self.w

    @property
    def y0(self) -> np.float64:
        """y0 position"""
        return self.y

    @property
    def y1(self) -> np.float64:
        """y1 position"""
        return self.y + self.h

    @classmethod
    def from_extent(cls, extent, canvas_rect):
        """create a RectManager from an extent"""
        rect = cls.extent_to_rect(extent, canvas_rect)
        return cls(*rect, canvas_rect)

    @property
    def extent(self) -> np.ndarray:
        """extent, (xmin, xmax, ymin, ymax)"""
        # not actually stored, computed when needed
        return np.asarray([self.x0, self.x1, self.y0, self.y1])

    @extent.setter
    def extent(self, extent):
        rect = RectManager.extent_to_rect(extent, canvas_rect=self._canvas_rect)

        self._set(rect)

    @staticmethod
    def extent_to_rect(extent, canvas_rect):
        """convert an extent to a rect"""
        RectManager.validate_extent(extent, canvas_rect)
        x0, x1, y0, y1 = extent

        # width and height
        w = x1 - x0
        h = y1 - y0

        return x0, y0, w, h

    @staticmethod
    def validate_extent(extent: np.ndarray | tuple, canvas_rect: tuple):
        extent = np.asarray(extent)
        cx0, cy0, cw, ch = canvas_rect

        # make sure extent is valid
        if (extent < 0).any():
            raise ValueError(f"extent must be non-negative, you have passed: {extent}")

        if extent[1] <= 1 or extent[3] <= 1:  # if x1 <= 1, or y1 <= 1
            # if fractional rect, convert to full
            if not (extent <= 1).all():  # if x1 and y1 <= 1, then all vals must be <= 1
                raise ValueError(
                    f"if passing a fractional extent, all values must be fractional, you have passed: {extent}"
                )
            extent *= np.asarray([cw, cw, ch, ch])

        x0, x1, y0, y1 = extent

        # width and height
        w = x1 - x0
        h = y1 - y0

        # check if x1 - x0 <= 0
        if w <= 0:
            raise ValueError(f"extent x-range must be non-negative: {extent}")

        # check if y1 - y0 <= 0
        if h <= 0:
            raise ValueError(f"extent y-range must be non-negative: {extent}")

        # calc canvas extent
        cx1 = cx0 + cw
        cy1 = cy0 + ch
        canvas_extent = np.asarray([cx0, cx1, cy0, cy1])

        if x0 < cx0 or x1 < cx0 or x0 > cx1 or x1 > cx1:
            raise ValueError(
                f"extent: {extent} x-range is beyond the bounds of the canvas: {canvas_extent}"
            )
        if y0 < cy0 or y1 < cy0 or y0 > cy1 or y1 > cy1:
            raise ValueError(
                f"extent: {extent} y-range is beyond the bounds of the canvas: {canvas_extent}"
            )

    def is_above(self, y0, dist: int = 1) -> bool:
        # our bottom < other top within given distance
        return self.y1 < y0 + dist

    def is_below(self, y1, dist: int = 1) -> bool:
        # our top > other bottom
        return self.y0 > y1 - dist

    def is_left_of(self, x0, dist: int = 1) -> bool:
        # our right_edge < other left_edge
        # self.x1 < other.x0
        return self.x1 < x0 + dist

    def is_right_of(self, x1, dist: int = 1) -> bool:
        # self.x0 > other.x1
        return self.x0 > x1 - dist

    def overlaps(self, extent: np.ndarray) -> bool:
        """returns whether this rect overlaps with the given extent"""
        x0, x1, y0, y1 = extent
        return not any(
            [
                self.is_above(y0),
                self.is_below(y1),
                self.is_left_of(x0),
                self.is_right_of(x1),
            ]
        )

    def __repr__(self):
        s = f"{self._rect_frac}\n{self.rect}"

        return s



from contextlib import contextmanager

from ._base import Graphic


@contextmanager
def pause_events(*graphics: Graphic):
    """
    Context manager for pausing Graphic events.

    Examples
    --------

    .. code-block::

        # pass in any number of graphics
        with fpl.pause_events(graphic1, graphic2, graphic3):
            # enter context manager
            # all events are blocked from graphic1, graphic2, graphic3

        # context manager exited, event states restored.

    """
    if not all([isinstance(g, Graphic) for g in graphics]):
        raise TypeError(
            f"`pause_events` only takes Graphic instances as arguments, "
            f"you have passed the following types:\n{[type(g) for g in graphics]}"
        )

    original_vals = [g.block_events for g in graphics]

    for g in graphics:
        g.block_events = True
    yield

    for g, value in zip(graphics, original_vals):
        g.block_events = value



from typing import Any

import numpy as np
import pygfx

from ...utils import (
    parse_cmap_values,
)
from ._base import (
    GraphicFeature,
    BufferManager,
    GraphicFeatureEvent,
    to_gpu_supported_dtype,
    block_reentrance,
)
from .utils import parse_colors


class VertexColors(BufferManager):
    property_name = "colors"
    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice, index, numpy-like fancy index",
            "description": "index/slice at which colors were indexed/sliced",
        },
        {
            "dict key": "value",
            "type": "np.ndarray [n_points_changed, RGBA]",
            "description": "new color values for points that were changed",
        },
        {
            "dict key": "user_value",
            "type": "str or array-like",
            "description": "user input value that was parsed into the RGBA array",
        },
    ]

    def __init__(
        self,
        colors: str | np.ndarray | tuple[float] | list[float] | list[str],
        n_colors: int,
        isolated_buffer: bool = True,
    ):
        """
        Manages the vertex color buffer for :class:`LineGraphic` or :class:`ScatterGraphic`

        Parameters
        ----------
        colors: str | np.ndarray | tuple[float, float, float, float] | list[str] | list[float] | int | float
            specify colors as a single human-readable string, RGBA array,
            or an iterable of strings or RGBA arrays

        n_colors: int
            number of colors, if passing in a single str or single RGBA array

        """
        data = parse_colors(colors, n_colors)

        super().__init__(data=data, isolated_buffer=isolated_buffer)

    @block_reentrance
    def __setitem__(
        self,
        key: int | slice | np.ndarray[int | bool] | tuple[slice, ...],
        user_value: str | np.ndarray | tuple[float] | list[float] | list[str],
    ):
        user_key = key

        if isinstance(key, tuple):
            # directly setting RGBA values for points, we do no parsing
            if not isinstance(user_value, (int, float, np.ndarray)):
                raise TypeError(
                    "Can only set from int, float, or array to set colors directly by slicing the entire array"
                )
            value = user_value

        elif isinstance(key, int):
            # set color of one point
            n_colors = 1
            value = parse_colors(user_value, n_colors)

        elif isinstance(key, slice):
            # find n_colors by converting slice to range and then parse colors
            start, stop, step = key.indices(self.value.shape[0])

            n_colors = len(range(start, stop, step))

            value = parse_colors(user_value, n_colors)

        elif isinstance(key, (np.ndarray, list)):
            if isinstance(key, list):
                # convert to array
                key = np.array(key)

            # make sure it's 1D
            if not key.ndim == 1:
                raise TypeError(
                    "If slicing colors with an array, it must be a 1D bool or int array"
                )

            if key.dtype == bool:
                # make sure len is same
                if not key.size == self.buffer.data.shape[0]:
                    raise IndexError(
                        f"Length of array for fancy indexing must match number of datapoints.\n"
                        f"There are {len(self.buffer.data.shape[0])} datapoints and you have passed {key.size} indices"
                    )
                n_colors = np.count_nonzero(key)

            elif np.issubdtype(key.dtype, np.integer):
                n_colors = key.size

            else:
                raise TypeError(
                    "If slicing colors with an array, it must be a 1D bool or int array"
                )

            value = parse_colors(user_value, n_colors)

        else:
            raise TypeError(
                f"invalid key for setting colors, you may set colors using integer indices, slices, or "
                f"fancy indexing using an array of integers or bool"
            )

        self.buffer.data[key] = value

        self._update_range(key)

        if len(self._event_handlers) < 1:
            return

        event_info = {
            "key": user_key,
            "value": value,
            "user_value": user_value,
        }

        event = GraphicFeatureEvent("colors", info=event_info)
        self._call_event_handlers(event)

    def __len__(self):
        return len(self.buffer.data)


class UniformColor(GraphicFeature):
    property_name = "colors"
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray [RGBA]",
            "description": "new color value",
        },
    ]

    def __init__(self, value: str | np.ndarray | tuple | list | pygfx.Color):
        """Manages uniform color for line or scatter material"""

        self._value = pygfx.Color(value)
        super().__init__()

    @property
    def value(self) -> pygfx.Color:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str | np.ndarray | tuple | list | pygfx.Color):
        value = pygfx.Color(value)
        graphic.world_object.material.color = value
        self._value = value

        event = GraphicFeatureEvent(type="colors", info={"value": value})
        self._call_event_handlers(event)


class UniformSize(GraphicFeature):
    property_name = "sizes"
    event_info_spec = [
        {"dict key": "value", "type": "float", "description": "new size value"},
    ]

    def __init__(self, value: int | float):
        """Manages uniform size for scatter material"""

        self._value = float(value)
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float | int):
        value = float(value)
        graphic.world_object.material.size = value
        self._value = value

        event = GraphicFeatureEvent(type="sizes", info={"value": value})
        self._call_event_handlers(event)


class SizeSpace(GraphicFeature):
    property_name = "size_space"
    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "'screen' | 'world' | 'model'",
        },
    ]

    def __init__(self, value: str):
        """Manages the coordinate space for scatter/line graphic"""

        self._value = value
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        if value not in ["screen", "world", "model"]:
            raise ValueError(
                f"`size_space` must be one of: {['screen', 'world', 'model']}"
            )

        if "Line" in graphic.world_object.material.__class__.__name__:
            graphic.world_object.material.thickness_space = value
        else:
            graphic.world_object.material.size_space = value
        self._value = value

        event = GraphicFeatureEvent(type="size_space", info={"value": value})
        self._call_event_handlers(event)


class VertexPositions(BufferManager):
    property_name = "data"
    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice, index (int) or numpy-like fancy index",
            "description": "key at which vertex positions data were indexed/sliced",
        },
        {
            "dict key": "value",
            "type": "int | float | array-like",
            "description": "new data values for points that were changed",
        },
    ]

    def __init__(self, data: Any, isolated_buffer: bool = True):
        """
        Manages the vertex positions buffer shown in the graphic.
        Supports fancy indexing if the data array also supports it.
        """

        data = self._fix_data(data)
        super().__init__(data, isolated_buffer=isolated_buffer)

    def _fix_data(self, data):
        # data = to_gpu_supported_dtype(data)

        if data.ndim == 1:
            # if user provides a 1D array, assume these are y-values
            data = np.column_stack([np.arange(data.size, dtype=data.dtype), data])

        if data.shape[1] != 3:
            if data.shape[1] != 2:
                raise ValueError(f"Must pass 1D, 2D or 3D data")

            # zeros for z
            zs = np.zeros(data.shape[0], dtype=data.dtype)

            # column stack [x, y, z] to make data of shape [n_points, 3]
            data = np.column_stack([data[:, 0], data[:, 1], zs])

        return to_gpu_supported_dtype(data)

    @block_reentrance
    def __setitem__(
        self,
        key: int | slice | np.ndarray[int | bool] | tuple[slice, ...],
        value: np.ndarray | float | list[float],
    ):
        # directly use the key to slice the buffer
        self.buffer.data[key] = value

        # _update_range handles parsing the key to
        # determine offset and size for GPU upload
        self._update_range(key)

        self._emit_event("data", key, value)

    def __len__(self):
        return len(self.buffer.data)


class PointsSizesFeature(BufferManager):
    property_name = "sizes"
    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice, index (int) or numpy-like fancy index",
            "description": "key at which point sizes were indexed/sliced",
        },
        {
            "dict key": "value",
            "type": "int | float | array-like",
            "description": "new size values for points that were changed",
        },
    ]

    def __init__(
        self,
        sizes: int | float | np.ndarray | list[int | float] | tuple[int | float],
        n_datapoints: int,
        isolated_buffer: bool = True,
    ):
        """
        Manages sizes buffer of scatter points.
        """
        sizes = self._fix_sizes(sizes, n_datapoints)
        super().__init__(data=sizes, isolated_buffer=isolated_buffer)

    def _fix_sizes(
        self,
        sizes: int | float | np.ndarray | list[int | float] | tuple[int | float],
        n_datapoints: int,
    ):
        if np.issubdtype(type(sizes), np.number):
            # single value given
            sizes = np.full(
                n_datapoints, sizes, dtype=np.float32
            )  # force it into a float to avoid weird gpu errors

        elif isinstance(
            sizes, (np.ndarray, tuple, list)
        ):  # if it's not a ndarray already, make it one
            sizes = np.asarray(sizes, dtype=np.float32)  # read it in as a numpy.float32
            if (sizes.ndim != 1) or (sizes.size != n_datapoints):
                raise ValueError(
                    f"sequence of `sizes` must be 1 dimensional with "
                    f"the same length as the number of datapoints"
                )

        else:
            raise TypeError(
                "sizes must be a single <int>, <float>, or a sequence (array, list, tuple) of int"
                "or float with the length equal to the number of datapoints"
            )

        if np.count_nonzero(sizes < 0) > 1:
            raise ValueError(
                "All sizes must be positive numbers greater than or equal to 0.0."
            )

        return sizes

    @block_reentrance
    def __setitem__(
        self,
        key: int | slice | np.ndarray[int | bool] | list[int | bool],
        value: int | float | np.ndarray | list[int | float] | tuple[int | float],
    ):
        # this is a very simple 1D buffer, no parsing required, directly set buffer
        self.buffer.data[key] = value
        self._update_range(key)

        self._emit_event("sizes", key, value)

    def __len__(self):
        return len(self.buffer.data)


class Thickness(GraphicFeature):
    property_name = "thickness"
    event_info_spec = [
        {"dict key": "value", "type": "float", "description": "new thickness value"},
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        value = float(value)
        graphic.world_object.material.thickness = value
        self._value = value

        event = GraphicFeatureEvent(type="thickness", info={"value": value})
        self._call_event_handlers(event)


class VertexCmap(BufferManager):
    property_name = "cmap"
    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice",
            "description": "key at cmap colors were sliced",
        },
        {
            "dict key": "value",
            "type": "str",
            "description": "new cmap to set at given slice",
        },
    ]

    def __init__(
        self,
        vertex_colors: VertexColors,
        cmap_name: str | None,
        transform: np.ndarray | None,
    ):
        """
        Sliceable colormap feature, manages a VertexColors instance and
        provides a way to set colormaps with arbitrary transforms
        """

        super().__init__(data=vertex_colors.buffer)

        self._vertex_colors = vertex_colors
        self._cmap_name = cmap_name
        self._transform = transform

        if self._cmap_name is not None:
            if not isinstance(self._cmap_name, str):
                raise TypeError(
                    f"cmap name must be of type <str>, you have passed: {self._cmap_name} of type: {type(self._cmap_name)}"
                )

            if self._transform is not None:
                self._transform = np.asarray(self._transform)

            n_datapoints = vertex_colors.value.shape[0]

            colors = parse_cmap_values(
                n_colors=n_datapoints,
                cmap_name=self._cmap_name,
                transform=self._transform,
            )
            # set vertex colors from cmap
            self._vertex_colors[:] = colors

    @block_reentrance
    def __setitem__(self, key: slice, cmap_name):
        if not isinstance(key, slice):
            raise TypeError(
                "fancy indexing not supported for VertexCmap, only slices "
                "of a continuous range are supported for applying a cmap"
            )
        if key.step is not None:
            raise TypeError(
                "step sized indexing not currently supported for setting VertexCmap, "
                "slices must be a continuous range"
            )

        # parse slice
        start, stop, step = key.indices(self.value.shape[0])
        n_elements = len(range(start, stop, step))

        colors = parse_cmap_values(
            n_colors=n_elements, cmap_name=cmap_name, transform=self._transform
        )

        self._cmap_name = cmap_name
        self._vertex_colors[key] = colors

        # TODO: should we block vertex_colors from emitting an event?
        #  Because currently this will result in 2 emitted events, one
        #  for cmap and another from the colors
        self._emit_event("cmap", key, cmap_name)

    @property
    def name(self) -> str:
        return self._cmap_name

    @property
    def transform(self) -> np.ndarray | None:
        """Get or set the cmap transform. Maps values from the transform array to the cmap colors"""
        return self._transform

    @transform.setter
    def transform(
        self,
        values: np.ndarray | list[float | int],
        indices: slice | list | np.ndarray = None,
    ):
        if self._cmap_name is None:
            raise AttributeError(
                "cmap name is not set, set the cmap name before setting the transform"
            )

        values = np.asarray(values)

        colors = parse_cmap_values(
            n_colors=self.value.shape[0], cmap_name=self._cmap_name, transform=values
        )

        self._transform = values

        if indices is None:
            indices = slice(None)

        self._vertex_colors[indices] = colors

        self._emit_event("cmap.transform", indices, values)

    def __len__(self):
        raise NotImplementedError(
            "len not implemented for `cmap`, use len(colors) instead"
        )

    def __repr__(self):
        return f"{self.__class__.__name__} | cmap: {self.name}\ntransform: {self.transform}"



from contextlib import suppress
from typing import Any

import numpy as np

from ._base import Graphic


class CollectionProperties:
    """
    Properties common to all Graphic Collections

    Allows getting and setting the common properties of the individual graphics in the collection
    """

    def _set_feature(self, feature, values):
        if not len(values) == len(self):
            raise IndexError

        for g, v in zip(self, values):
            setattr(g, feature, v)

    @property
    def names(self) -> np.ndarray[str | None]:
        """get or set the name of the individual graphics in the collection"""
        return np.asarray([g.name for g in self])

    @names.setter
    def names(self, values: np.ndarray[str] | list[str]):
        self._set_feature("name", values)

    @property
    def metadatas(self) -> np.ndarray[str | None]:
        """get or set the metadata of the individual graphics in the collection"""
        return np.asarray([g.metadata for g in self])

    @metadatas.setter
    def metadatas(self, values: np.ndarray[str] | list[str]):
        self._set_feature("metadata", values)

    @property
    def offsets(self) -> np.ndarray:
        """get or set the offset of the individual graphics in the collection"""
        return np.stack([g.offset for g in self])

    @offsets.setter
    def offsets(self, values: np.ndarray | list[np.ndarray]):
        self._set_feature("offset", values)

    @property
    def rotations(self) -> np.ndarray:
        """get or set the rotation of the individual graphics in the collection"""
        return np.stack([g.rotation for g in self])

    @rotations.setter
    def rotations(self, values: np.ndarray | list[np.ndarray]):
        self._set_feature("rotation", values)

    # TODO: how to work with deleted feature in a collection

    @property
    def visibles(self) -> np.ndarray[bool]:
        """get or set the offsets of the individual graphics in the collection"""
        return np.asarray([g.visible for g in self])

    @visibles.setter
    def visibles(self, values: np.ndarray[bool] | list[bool]):
        self._set_feature("visible", values)


class CollectionIndexer(CollectionProperties):
    """Collection Indexer"""

    def __init__(self, selection: np.ndarray[Graphic], features: set[str]):
        """

        Parameters
        ----------

        selection: np.ndarray of Graphics
            array of the selected Graphics from the parent GraphicCollection based on the ``selection_indices``

        """

        if isinstance(selection, Graphic):
            selection = np.asarray([selection])

        self._selection = selection
        self.features = features

    @property
    def graphics(self) -> np.ndarray[Graphic]:
        """Returns an array of the selected graphics"""
        return tuple(self._selection)

    def add_event_handler(self, *args):
        """
        Register an event handler.

        Parameters
        ----------
        callback: callable, the first argument
            Event handler, must accept a single event  argument
        *types: list of strings
            A list of event types, ex: "click", "data", "colors", "pointer_down"

        For the available renderer event types, see
        https://jupyter-rfb.readthedocs.io/en/stable/events.html

        All feature support events, i.e. ``graphic.features`` will give a set of
        all features that are evented

        Can also be used as a decorator.

        Example
        -------

        .. code-block:: py

            def my_handler(event):
                print(event)

            graphic.add_event_handler(my_handler, "pointer_up", "pointer_down")

        Decorator usage example:

        .. code-block:: py

            @graphic.add_event_handler("click")
            def my_handler(event):
                print(event)
        """

        decorating = not callable(args[0])
        types = args if decorating else args[1:]

        if decorating:

            def decorator(_callback):
                for g in self:
                    g.add_event_handler(_callback, *types)
                return _callback

            return decorator

        for g in self:
            g.add_event_handler(*args)

    def remove_event_handler(self, callback, *types):
        for g in self:
            g.remove_event_handler(callback, *types)

    def clear_event_handlers(self):
        for g in self:
            g.clear_event_handlers()

    def __getitem__(self, item):
        return self.graphics[item]

    def __len__(self):
        return len(self._selection)

    def __iter__(self):
        self._iter = iter(range(len(self)))
        return self

    def __next__(self) -> Graphic:
        index = next(self._iter)

        return self.graphics[index]

    def __repr__(self):
        return (
            f"{self.__class__.__name__} @ {hex(id(self))}\n"
            f"Selection of <{len(self._selection)}> {self._selection[0].__class__.__name__}"
        )


class GraphicCollection(Graphic, CollectionProperties):
    """Graphic Collection base class"""

    _child_type: type
    _indexer: type

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._features = cls._child_type._features

    def __init__(self, name: str = None, metadata: Any = None, **kwargs):
        super().__init__(name=name, metadata=metadata, **kwargs)

        # list of mem locations of the graphics
        self._graphics: list[Graphic] = list()

        self._graphics_changed: bool = True

        self._iter = None

    @property
    def graphics(self) -> np.ndarray[Graphic]:
        """The Graphics within this collection."""

        return np.asarray(self._graphics)

    def add_graphic(self, graphic: Graphic):
        """
        Add a graphic to the collection.

        Parameters
        ----------
        graphic: Graphic
            graphic to add, must be a real ``Graphic`` not a proxy

        """

        if not type(graphic) == self._child_type:
            raise TypeError(
                f"Can only add graphics of the same type to a collection.\n"
                f"You can only add {self._child_type.__name__} to a {self.__class__.__name__}, "
                f"you are trying to add a {graphic.__class__.__name__}."
            )

        self._graphics.append(graphic)

        self.world_object.add(graphic.world_object)

        self._graphics_changed = True

    def remove_graphic(self, graphic: Graphic):
        """
        Remove a graphic from the collection.

        Note: Only removes the graphic from the collection. Does not remove
        the graphic from the scene, and does not delete the graphic.

        Parameters
        ----------
        graphic: Graphic
            graphic to remove

        """

        self._graphics.remove(graphic)

        self.world_object.remove(graphic.world_object)

        self._graphics_changed = True

    def add_event_handler(self, *args):
        """
        Register an event handler.

        Parameters
        ----------
        callback: callable, the first argument
            Event handler, must accept a single event  argument
        *types: list of strings
            A list of event types, ex: "click", "data", "colors", "pointer_down"

        For the available renderer event types, see
        https://jupyter-rfb.readthedocs.io/en/stable/events.html

        All feature support events, i.e. ``graphic.features`` will give a set of
        all features that are evented

        Can also be used as a decorator.

        Example
        -------

        .. code-block:: py

            def my_handler(event):
                print(event)

            graphic.add_event_handler(my_handler, "pointer_up", "pointer_down")

        Decorator usage example:

        .. code-block:: py

            @graphic.add_event_handler("click")
            def my_handler(event):
                print(event)
        """

        return self[:].add_event_handler(*args)

    def remove_event_handler(self, callback, *types):
        """remove an event handler"""
        self[:].remove_event_handler(callback, *types)

    def clear_event_handlers(self):
        self[:].clear_event_handlers()

    def _fpl_add_plot_area_hook(self, plot_area):
        super()._fpl_add_plot_area_hook(plot_area)

        for g in self:
            g._fpl_add_plot_area_hook(plot_area)

    def _fpl_prepare_del(self):
        """
        Cleans up the graphic in preparation for __del__(), such as removing event handlers from
        plot renderer, feature event handlers, etc.

        Optionally implemented in subclasses
        """
        # clear any attached event handlers and animation functions
        self.world_object._event_handlers.clear()

        for g in self:
            g._fpl_prepare_del()

    def __getitem__(self, key) -> CollectionIndexer:
        if np.issubdtype(type(key), np.integer):
            return self.graphics[key]

        return self._indexer(selection=self.graphics[key], features=self._features)

    def __del__(self):
        # detach children
        self.world_object.clear()

        for g in self.graphics:
            g._fpl_prepare_del()
            del g

        super().__del__()

    def __len__(self):
        return len(self._graphics)

    def __iter__(self):
        self._iter = iter(range(len(self)))
        return self

    def __next__(self) -> Graphic:
        index = next(self._iter)

        return self._graphics[index]

    def __repr__(self):
        rval = super().__repr__()
        return f"{rval}\nCollection of <{len(self._graphics)}> Graphics"


class CollectionFeature:
    """Collection Feature"""

    def __init__(self, selection: np.ndarray[Graphic], feature: str):
        """
        selection: list of Graphics
            a list of the selected Graphics from the parent GraphicCollection based on the ``selection_indices``

        feature: str
            feature of Graphics in the GraphicCollection being indexed

        """

        self._selection = selection
        self._feature = feature

        self._feature_instances = [getattr(g, feature) for g in self._selection]

    def __getitem__(self, item):
        return np.stack([fi[item] for fi in self._feature_instances])

    def __setitem__(self, key, value):
        for fi in self._feature_instances:
            fi[key] = value

    def __repr__(self):
        return f"Collection feature for: <{self._feature}>"



from typing import *

import numpy as np

import pygfx

from ..utils import parse_cmap_values
from ._collection_base import CollectionIndexer, GraphicCollection, CollectionFeature
from .line import LineGraphic
from .selectors import (
    LinearRegionSelector,
    LinearSelector,
    RectangleSelector,
    PolygonSelector,
)


class _LineCollectionProperties:
    """Mix-in class for LineCollection properties"""

    @property
    def colors(self) -> CollectionFeature:
        """get or set colors of lines in the collection"""
        return CollectionFeature(self.graphics, "colors")

    @colors.setter
    def colors(self, values: str | np.ndarray | tuple[float] | list[float] | list[str]):
        if isinstance(values, str):
            # set colors of all lines to one str color
            for g in self:
                g.colors = values
            return

        elif all(isinstance(v, str) for v in values):
            # individual str colors for each line
            if not len(values) == len(self):
                raise IndexError

            for g, v in zip(self.graphics, values):
                g.colors = v

            return

        if isinstance(values, np.ndarray):
            if values.ndim == 2:
                # assume individual colors for each
                for g, v in zip(self, values):
                    g.colors = v
                return

        elif len(values) == 4:
            # assume RGBA
            self.colors[:] = values

        else:
            # assume individual colors for each
            for g, v in zip(self, values):
                g.colors = v

    @property
    def data(self) -> CollectionFeature:
        """get or set data of lines in the collection"""
        return CollectionFeature(self.graphics, "data")

    @data.setter
    def data(self, values):
        for g, v in zip(self, values):
            g.data = v

    @property
    def cmap(self) -> CollectionFeature:
        """
        Get or set a cmap along the line collection.

        Optionally set using a tuple ("cmap", <transform>) to set the transform..
        Example:

        line_collection.cmap = ("jet", sine_transform_vals, 0.7)

        """
        return CollectionFeature(self.graphics, "cmap")

    @cmap.setter
    def cmap(self, args):
        if isinstance(args, str):
            name = args
            transform = None
        elif len(args) == 1:
            name = args[0]
            transform = None
        elif len(args) == 2:
            name, transform = args
        else:
            raise ValueError(
                "Too many values for cmap (note that alpha is deprecated, set alpha on the graphic instead)"
            )

        self.colors = parse_cmap_values(
            n_colors=len(self), cmap_name=name, transform=transform
        )

    @property
    def thickness(self) -> np.ndarray:
        """get or set the thickness of the lines"""
        return np.asarray([g.thickness for g in self])

    @thickness.setter
    def thickness(self, values: np.ndarray | list[float]):
        if not len(values) == len(self):
            raise IndexError

        for g, v in zip(self, values):
            g.thickness = v


class LineCollectionIndexer(CollectionIndexer, _LineCollectionProperties):
    """Indexer for line collections"""

    pass


class LineCollection(GraphicCollection, _LineCollectionProperties):
    _child_type = LineGraphic
    _indexer = LineCollectionIndexer

    def __init__(
        self,
        data: np.ndarray | List[np.ndarray],
        thickness: float | Sequence[float] = 2.0,
        colors: str | Sequence[str] | np.ndarray | Sequence[np.ndarray] = "w",
        uniform_colors: bool = False,
        cmap: Sequence[str] | str = None,
        cmap_transform: np.ndarray | List = None,
        name: str = None,
        names: list[str] = None,
        metadata: Any = None,
        metadatas: Sequence[Any] | np.ndarray = None,
        isolated_buffer: bool = True,
        kwargs_lines: list[dict] = None,
        **kwargs,
    ):
        """
        Create a collection of :class:`.LineGraphic`

        Parameters
        ----------
        data: list of array-like
            List or array-like of multiple line data to plot

            | if ``list`` each item in the list must be a 1D, 2D, or 3D numpy array
            | if  array-like, must be of shape [n_lines, n_points_line, y | xy | xyz]

        thickness: float or Iterable of float, default 2.0
            | if ``float``, single thickness will be used for all lines
            | if ``list`` of ``float``, each value will apply to the individual lines

        colors: str, RGBA array, Iterable of RGBA array, or Iterable of str, default "w"
            | if single ``str`` such as "w", "r", "b", etc, represents a single color for all lines
            | if single ``RGBA array`` (tuple or list of size 4), represents a single color for all lines
            | if ``list`` of ``str``, represents color for each individual line, example ["w", "b", "r",...]
            | if ``RGBA array`` of shape [data_size, 4], represents a single RGBA array for each line

        cmap: Iterable of str or str, optional
            | if ``str``, single cmap will be used for all lines
            | if ``list`` of ``str``, each cmap will apply to the individual lines

            .. note::
                ``cmap`` overrides any arguments passed to ``colors``

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        name: str, optional
            name of the line collection as a whole

        names: list[str], optional
            names of the individual lines in the collection, ``len(names)`` must equal ``len(data)``

        metadata: Any
            meatadata associated with the collection as a whole

        metadatas: Iterable or array
            metadata for each individual line associated with this collection, this is for the user to manage.
            ``len(metadata)`` must be same as ``len(data)``

        kwargs_lines: list[dict], optional
            list of kwargs passed to the individual lines, ``len(kwargs_lines)`` must equal ``len(data)``

        kwargs_collection
            kwargs for the collection, passed to GraphicCollection

        """

        super().__init__(name=name, metadata=metadata, **kwargs)

        if not isinstance(thickness, (float, int)):
            if len(thickness) != len(data):
                raise ValueError(
                    f"len(thickness) != len(data)\n{len(thickness)} != {len(data)}"
                )

        if names is not None:
            if len(names) != len(data):
                raise ValueError(
                    f"len(names) != len(data)\n{len(names)} != {len(data)}"
                )

        if metadatas is not None:
            if len(metadatas) != len(data):
                raise ValueError(
                    f"len(metadata) != len(data)\n{len(metadatas)} != {len(data)}"
                )

        if kwargs_lines is not None:
            if len(kwargs_lines) != len(data):
                raise ValueError(
                    f"len(kwargs_lines) != len(data)\n"
                    f"{len(kwargs_lines)} != {len(data)}"
                )

        self._cmap_transform = cmap_transform
        self._cmap_str = cmap

        # cmap takes priority over colors
        if cmap is not None:
            # cmap across lines
            if isinstance(cmap, str):
                colors = parse_cmap_values(
                    n_colors=len(data), cmap_name=cmap, transform=cmap_transform
                )
                single_color = False
                cmap = None

            elif isinstance(cmap, (tuple, list)):
                if len(cmap) != len(data):
                    raise ValueError(
                        "cmap argument must be a single cmap or a list of cmaps "
                        "with the same length as the data"
                    )
                single_color = False
            else:
                raise ValueError(
                    "cmap argument must be a single cmap or a list of cmaps "
                    "with the same length as the data"
                )
        else:
            if isinstance(colors, np.ndarray):
                # single color for all lines in the collection as RGBA
                if colors.shape in [(3,), (4,)]:
                    single_color = True

                # colors specified for each line as array of shape [n_lines, RGBA]
                elif colors.shape == (len(data), 4):
                    single_color = False

                else:
                    raise ValueError(
                        f"numpy array colors argument must be of shape (4,) or (n_lines, 4)."
                        f"You have pass the following shape: {colors.shape}"
                    )

            elif isinstance(colors, str):
                if colors == "random":
                    colors = np.random.rand(len(data), 3)
                    single_color = False
                else:
                    # parse string color
                    single_color = True
                    colors = pygfx.Color(colors)

            elif isinstance(colors, (tuple, list)):
                if len(colors) == 4:
                    # single color specified as (R, G, B, A) tuple or list
                    if all([isinstance(c, (float, int)) for c in colors]):
                        single_color = True

                elif len(colors) == len(data):
                    # colors passed as list/tuple of colors, such as list of string
                    single_color = False

                else:
                    raise ValueError(
                        "tuple or list colors argument must be a single color represented as [R, G, B, A], "
                        "or must be a tuple/list of colors represented by a string with the same length as the data"
                    )

        if kwargs_lines is None:
            kwargs_lines = dict()

        self._set_world_object(pygfx.Group())

        for i, d in enumerate(data):
            if isinstance(thickness, list):
                _s = thickness[i]
            else:
                _s = thickness

            if cmap is None:
                _cmap = None

                if single_color:
                    _c = colors
                else:
                    _c = colors[i]
            else:
                _cmap = cmap[i]
                _c = None

            if metadatas is not None:
                _m = metadatas[i]
            else:
                _m = None

            if names is not None:
                _name = names[i]
            else:
                _name = None

            lg = LineGraphic(
                data=d,
                thickness=_s,
                colors=_c,
                uniform_color=uniform_colors,
                cmap=_cmap,
                name=_name,
                metadata=_m,
                isolated_buffer=isolated_buffer,
                **kwargs_lines,
            )

            self.add_graphic(lg)

    def __getitem__(self, item) -> LineCollectionIndexer:
        return super().__getitem__(item)

    def add_linear_selector(
        self, selection: float = None, padding: float = 0.0, axis: str = "x", **kwargs
    ) -> LinearSelector:
        """
        Adds a linear selector.

        Parameters
        ----------
        Parameters
        ----------
        selection: float, optional
            selected point on the linear selector, computed from data if not provided

        axis: str, default "x"
            axis that the selector resides on

        padding: float, default 0.0
            Extra padding to extend the linear selector along the orthogonal axis to make it easier to interact with.

        kwargs
            passed to :class:`.LinearSelector`

        Returns
        -------
        LinearSelector

        """

        bounds_init, limits, size, center = self._get_linear_selector_init_args(
            axis, padding
        )

        if selection is None:
            selection = bounds_init[0]

        selector = LinearSelector(
            selection=selection,
            limits=limits,
            axis=axis,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_linear_region_selector(
        self,
        selection: tuple[float, float] = None,
        padding: float = 0.0,
        axis: str = "x",
        **kwargs,
    ) -> LinearRegionSelector:
        """
        Add a :class:`.LinearRegionSelector`. Selectors are just ``Graphic`` objects, so you can manage,
        remove, or delete them from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float), optional
            the starting bounds of the linear region selector, computed from data if not provided

        axis: str, default "x"
            axis that the selector resides on

        padding: float, default 0.0
            Extra padding to extend the linear region selector along the orthogonal axis to make it easier to interact with.

        kwargs
            passed to ``LinearRegionSelector``

        Returns
        -------
        LinearRegionSelector
            linear selection graphic

        """

        bounds_init, limits, size, center = self._get_linear_selector_init_args(
            axis, padding
        )

        if selection is None:
            selection = bounds_init

        # create selector
        selector = LinearRegionSelector(
            selection=selection,
            limits=limits,
            size=size,
            center=center,
            axis=axis,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        # PlotArea manages this for garbage collection etc. just like all other Graphics
        # so we should only work with a proxy on the user-end
        return selector

    def add_rectangle_selector(
        self,
        selection: tuple[float, float, float] = None,
        **kwargs,
    ) -> RectangleSelector:
        """
        Add a :class:`.RectangleSelector`. Selectors are just ``Graphic`` objects, so you can manage,
        remove, or delete them from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float, float, float), optional
            initial (xmin, xmax, ymin, ymax) of the selection
        """
        bbox = self.world_object.get_world_bounding_box()

        xdata = np.array(self.data[:, 0])
        xmin, xmax = (np.nanmin(xdata), np.nanmax(xdata))
        value_25px = (xmax - xmin) / 4

        ydata = np.array(self.data[:, 1])
        ymin = np.floor(ydata.min()).astype(int)

        ymax = np.ptp(bbox[:, 1])

        if selection is None:
            selection = (xmin, value_25px, ymin, ymax)

        limits = (xmin, xmax, ymin - (ymax * 1.5 - ymax), ymax * 1.5)

        selector = RectangleSelector(
            selection=selection,
            limits=limits,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_polygon_selector(
        self,
        selection: List[tuple[float, float]] = None,
        **kwargs,
    ) -> PolygonSelector:
        """
        Add a :class:`.PolygonSelector`. Selectors are just ``Graphic`` objects, so you can manage,
        remove, or delete them from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: List of positions, optional
            Initial points for the polygon. If not given or None, you'll start drawing the selection (clicking adds points to the polygon).
        """
        bbox = self.world_object.get_world_bounding_box()

        xdata = np.array(self.data[:, 0])
        xmin, xmax = (np.nanmin(xdata), np.nanmax(xdata))

        ydata = np.array(self.data[:, 1])
        ymin = np.floor(ydata.min()).astype(int)

        ymax = np.ptp(bbox[:, 1])

        limits = (xmin, xmax, ymin - (ymax * 1.5 - ymax), ymax * 1.5)

        selector = PolygonSelector(
            selection,
            limits,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def _get_linear_selector_init_args(self, axis, padding):
        # use bbox to get size and center
        bbox = self.world_object.get_world_bounding_box()

        if axis == "x":
            xdata = np.array(self.data[:, 0])
            xmin, xmax = (np.nanmin(xdata), np.nanmax(xdata))
            value_25p = (xmax - xmin) / 4

            bounds = (xmin, value_25p)
            limits = (xmin, xmax)
            # size from orthogonal axis
            size = np.ptp(bbox[:, 1]) * 1.5
            # center on orthogonal axis
            center = bbox[:, 1].mean()

        elif axis == "y":
            ydata = np.array(self.data[:, 1])
            xmin, xmax = (np.nanmin(ydata), np.nanmax(ydata))
            value_25p = (xmax - xmin) / 4

            bounds = (xmin, value_25p)
            limits = (xmin, xmax)

            size = np.ptp(bbox[:, 0]) * 1.5
            # center on orthogonal axis
            center = bbox[:, 0].mean()

        return bounds, limits, size, center


axes = {"x": 0, "y": 1, "z": 2}


class LineStack(LineCollection):
    def __init__(
        self,
        data: List[np.ndarray],
        thickness: float | Iterable[float] = 2.0,
        colors: str | Iterable[str] | np.ndarray | Iterable[np.ndarray] = "w",
        cmap: Iterable[str] | str = None,
        cmap_transform: np.ndarray | List = None,
        name: str = None,
        names: list[str] = None,
        metadata: Any = None,
        metadatas: Sequence[Any] | np.ndarray = None,
        isolated_buffer: bool = True,
        separation: float = 10.0,
        separation_axis: str = "y",
        kwargs_lines: list[dict] = None,
        **kwargs,
    ):
        """
        Create a stack of :class:`.LineGraphic` that are separated along the "x" or "y" axis.

        Parameters
        ----------
        data: list of array-like
            List or array-like of multiple line data to plot

            | if ``list`` each item in the list must be a 1D, 2D, or 3D numpy array
            | if  array-like, must be of shape [n_lines, n_points_line, y | xy | xyz]

        thickness: float or Iterable of float, default 2.0
            | if ``float``, single thickness will be used for all lines
            | if ``list`` of ``float``, each value will apply to the individual lines

        colors: str, RGBA array, Iterable of RGBA array, or Iterable of str, default "w"
            | if single ``str`` such as "w", "r", "b", etc, represents a single color for all lines
            | if single ``RGBA array`` (tuple or list of size 4), represents a single color for all lines
            | if ``list`` of ``str``, represents color for each individual line, example ["w", "b", "r",...]
            | if ``RGBA array`` of shape [data_size, 4], represents a single RGBA array for each line

        cmap: Iterable of str or str, optional
            | if ``str``, single cmap will be used for all lines
            | if ``list`` of ``str``, each cmap will apply to the individual lines

            .. note::
                ``cmap`` overrides any arguments passed to ``colors``

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        name: str, optional
            name of the line collection as a whole

        names: list[str], optional
            names of the individual lines in the collection, ``len(names)`` must equal ``len(data)``

        metadata: Any
            metadata associated with the collection as a whole

        metadatas: Iterable or array
            metadata for each individual line associated with this collection, this is for the user to manage.
            ``len(metadata)`` must be same as ``len(data)``

        separation: float, default 10
            space in between each line graphic in the stack

        separation_axis: str, default "y"
            axis in which the line graphics in the stack should be separated


        kwargs_lines: list[dict], optional
            list of kwargs passed to the individual lines, ``len(kwargs_lines)`` must equal ``len(data)``

        kwargs_collection
            kwargs for the collection, passed to GraphicCollection

        """
        super().__init__(
            data=data,
            thickness=thickness,
            colors=colors,
            cmap=cmap,
            cmap_transform=cmap_transform,
            name=name,
            names=names,
            metadata=metadata,
            metadatas=metadatas,
            isolated_buffer=isolated_buffer,
            kwargs_lines=kwargs_lines,
            **kwargs,
        )

        axis_zero = 0
        for i, line in enumerate(self.graphics):
            if separation_axis == "x":
                line.offset = (axis_zero, *line.offset[1:])

            elif separation_axis == "y":
                line.offset = (line.offset[0], axis_zero, line.offset[2])

            axis_zero = (
                axis_zero + line.data.value[:, axes[separation_axis]].max() + separation
            )

        self.separation = separation



from math import ceil
from typing import Sequence
import weakref

import numpy as np

import pygfx

from ..utils import subsample_array
from ..graphics import LineGraphic, ImageGraphic, ImageVolumeGraphic, TextGraphic
from ..graphics.utils import pause_events
from ..graphics._base import Graphic
from ..graphics.selectors import LinearRegionSelector


def _get_image_graphic_events(image_graphic: ImageGraphic) -> list[str]:
    """Small helper function to return the relevant events for an ImageGraphic"""
    events = ["vmin", "vmax"]

    if not image_graphic.data.value.ndim > 2:
        events.append("cmap")

    # if RGB(A), do not add cmap

    return events


# TODO: This is a widget, we can think about a BaseWidget class later if necessary
class HistogramLUTTool(Graphic):
    def __init__(
        self,
        data: np.ndarray,
        images: (
            ImageGraphic
            | ImageVolumeGraphic
            | Sequence[ImageGraphic | ImageVolumeGraphic]
        ),
        nbins: int = 100,
        flank_divisor: float = 5.0,
        **kwargs,
    ):
        """
        HistogramLUT tool that can be used to control the vmin, vmax of ImageGraphics or ImageVolumeGraphics.
        If used to control multiple images or image volumes it is assumed that they share a representation of
        the same data, and that their histogram, vmin, and vmax are identical. For example, displaying a
        ImageVolumeGraphic and several images that represent slices of the same volume data.

        Parameters
        ----------
        data: np.ndarray

        images: ImageGraphic | ImageVolumeGraphic | tuple[ImageGraphic | ImageVolumeGraphic]

        nbins: int, defaut 100.
            Total number of bins used in the histogram

        flank_divisor: float, default 5.0.
            Fraction of empty histogram bins on the tails of the distribution set `np.inf` for no flanks

        kwargs: passed to ``Graphic``

        """
        super().__init__(**kwargs)

        self._nbins = nbins
        self._flank_divisor = flank_divisor

        if isinstance(images, (ImageGraphic, ImageVolumeGraphic)):
            images = (images,)
        elif isinstance(images, Sequence):
            if not all(
                [isinstance(ig, (ImageGraphic, ImageVolumeGraphic)) for ig in images]
            ):
                raise TypeError(
                    f"`images` argument must be an ImageGraphic, ImageVolumeGraphic, or a "
                    f"tuple or list or ImageGraphic | ImageVolumeGraphic"
                )
        else:
            raise TypeError(
                f"`images` argument must be an ImageGraphic, ImageVolumeGraphic, or a "
                f"tuple or list or ImageGraphic | ImageVolumeGraphic"
            )

        self._images = images

        self._data = weakref.proxy(data)

        self._scale_factor: float = 1.0

        hist, edges, hist_scaled, edges_flanked = self._calculate_histogram(data)

        line_data = np.column_stack([hist_scaled, edges_flanked])

        self._histogram_line = LineGraphic(
            line_data, colors=(0.8, 0.8, 0.8), alpha_mode="solid", offset=(0, 0, -1)
        )

        bounds = (edges[0] * self._scale_factor, edges[-1] * self._scale_factor)
        limits = (edges_flanked[0], edges_flanked[-1])
        size = 120  # since it's scaled to 100
        origin = (hist_scaled.max() / 2, 0)

        self._linear_region_selector = LinearRegionSelector(
            selection=bounds,
            limits=limits,
            size=size,
            center=origin[0],
            axis="y",
            edge_thickness=8,
            parent=self._histogram_line,
        )

        self._vmin = self.images[0].vmin
        self._vmax = self.images[0].vmax

        # there will be a small difference with the histogram edges so this makes them both line up exactly
        self._linear_region_selector.selection = (
            self._vmin * self._scale_factor,
            self._vmax * self._scale_factor,
        )

        vmin_str, vmax_str = self._get_vmin_vmax_str()

        self._text_vmin = TextGraphic(
            text=vmin_str,
            font_size=16,
            offset=(0, 0, 0),
            anchor="top-left",
            outline_color="black",
            outline_thickness=0.5,
            alpha_mode="solid",
        )

        self._text_vmin.world_object.material.pick_write = False

        self._text_vmax = TextGraphic(
            text=vmax_str,
            font_size=16,
            offset=(0, 0, 0),
            anchor="bottom-left",
            outline_color="black",
            outline_thickness=0.5,
            alpha_mode="solid",
        )

        self._text_vmax.world_object.material.pick_write = False

        widget_wo = pygfx.Group()
        widget_wo.add(
            self._histogram_line.world_object,
            self._linear_region_selector.world_object,
            self._text_vmin.world_object,
            self._text_vmax.world_object,
        )

        self._set_world_object(widget_wo)

        self.world_object.local.scale_x *= -1

        self._text_vmin.offset = (-120, self._linear_region_selector.selection[0], 0)

        self._text_vmax.offset = (-120, self._linear_region_selector.selection[1], 0)

        self._linear_region_selector.add_event_handler(
            self._linear_region_handler, "selection"
        )

        ig_events = _get_image_graphic_events(self.images[0])

        for ig in self.images:
            ig.add_event_handler(self._image_cmap_handler, *ig_events)

        # colorbar for grayscale images
        if self.images[0].cmap is not None:
            self._colorbar: ImageGraphic = self._make_colorbar(edges_flanked)
            self._colorbar.add_event_handler(self._open_cmap_picker, "click")

            self.world_object.add(self._colorbar.world_object)
        else:
            self._colorbar = None
            self._cmap = None

    def _make_colorbar(self, edges_flanked) -> ImageGraphic:
        # use the histogram edge values as data for an
        # image with 2 columns, this will be our colorbar!
        colorbar_data = np.column_stack(
            [
                np.linspace(
                    edges_flanked[0], edges_flanked[-1], ceil(np.ptp(edges_flanked))
                )
            ]
            * 2
        ).astype(np.float32)

        colorbar_data /= self._scale_factor

        cbar = ImageGraphic(
            data=colorbar_data,
            vmin=self.vmin,
            vmax=self.vmax,
            cmap=self.images[0].cmap,
            interpolation="linear",
            offset=(-55, edges_flanked[0], -1),
        )

        cbar.world_object.world.scale_x = 20
        self._cmap = self.images[0].cmap

        return cbar

    def _get_vmin_vmax_str(self) -> tuple[str, str]:
        if self.vmin < 0.001 or self.vmin > 99_999:
            vmin_str = f"{self.vmin:.2e}"
        else:
            vmin_str = f"{self.vmin:.2f}"

        if self.vmax < 0.001 or self.vmax > 99_999:
            vmax_str = f"{self.vmax:.2e}"
        else:
            vmax_str = f"{self.vmax:.2f}"

        return vmin_str, vmax_str

    def _fpl_add_plot_area_hook(self, plot_area):
        self._plot_area = plot_area
        self._linear_region_selector._fpl_add_plot_area_hook(plot_area)
        self._histogram_line._fpl_add_plot_area_hook(plot_area)

        self._plot_area.auto_scale()
        self._plot_area.controller.enabled = True

    def _calculate_histogram(self, data):

        # get a subsampled view of this array
        data_ss = subsample_array(data, max_size=int(1e6))  # 1e6 is default
        hist, edges = np.histogram(data_ss, bins=self._nbins)

        # used if data ptp <= 10 because event things get weird
        # with tiny world objects due to floating point error
        # so if ptp <= 10, scale up by a factor
        data_interval = edges[-1] - edges[0]
        self._scale_factor: int = max(1, 100 * int(10 / data_interval))

        edges = edges * self._scale_factor

        bin_width = edges[1] - edges[0]

        flank_nbins = int(self._nbins / self._flank_divisor)
        flank_size = flank_nbins * bin_width

        flank_left = np.arange(edges[0] - flank_size, edges[0], bin_width)
        flank_right = np.arange(
            edges[-1] + bin_width, edges[-1] + flank_size, bin_width
        )

        edges_flanked = np.concatenate((flank_left, edges, flank_right))

        hist_flanked = np.concatenate(
            (np.zeros(flank_nbins), hist, np.zeros(flank_nbins))
        )

        # scale 0-100 to make it easier to see
        # float32 data can produce unnecessarily high values
        hist_scale_value = hist_flanked.max()
        if np.allclose(hist_scale_value, 0):
            hist_scale_value = 1
        hist_scaled = hist_flanked / (hist_scale_value / 100)

        if edges_flanked.size > hist_scaled.size:
            # we don't care about accuracy here so if it's off by 1-2 bins that's fine
            edges_flanked = edges_flanked[: hist_scaled.size]

        return hist, edges, hist_scaled, edges_flanked

    def _linear_region_handler(self, ev):
        # must use world coordinate values directly from selection()
        # otherwise the linear region bounds jump to the closest bin edges
        selected_ixs = self._linear_region_selector.selection
        vmin, vmax = selected_ixs[0], selected_ixs[1]
        vmin, vmax = vmin / self._scale_factor, vmax / self._scale_factor
        self.vmin, self.vmax = vmin, vmax

    def _image_cmap_handler(self, ev):
        setattr(self, ev.type, ev.info["value"])

    @property
    def cmap(self) -> str:
        return self._cmap

    @cmap.setter
    def cmap(self, name: str):
        if self._colorbar is None:
            return

        with pause_events(*self.images):
            for ig in self.images:
                ig.cmap = name

            self._cmap = name
            self._colorbar.cmap = name

    @property
    def vmin(self) -> float:
        return self._vmin

    @vmin.setter
    def vmin(self, value: float):
        with pause_events(self._linear_region_selector, *self.images):
            # must use world coordinate values directly from selection()
            # otherwise the linear region bounds jump to the closest bin edges
            self._linear_region_selector.selection = (
                value * self._scale_factor,
                self._linear_region_selector.selection[1],
            )
            for ig in self.images:
                ig.vmin = value

        self._vmin = value
        if self._colorbar is not None:
            self._colorbar.vmin = value

        vmin_str, vmax_str = self._get_vmin_vmax_str()
        self._text_vmin.offset = (-120, self._linear_region_selector.selection[0], 0)
        self._text_vmin.text = vmin_str

    @property
    def vmax(self) -> float:
        return self._vmax

    @vmax.setter
    def vmax(self, value: float):
        with pause_events(self._linear_region_selector, *self.images):
            # must use world coordinate values directly from selection()
            # otherwise the linear region bounds jump to the closest bin edges
            self._linear_region_selector.selection = (
                self._linear_region_selector.selection[0],
                value * self._scale_factor,
            )

            for ig in self.images:
                ig.vmax = value

        self._vmax = value
        if self._colorbar is not None:
            self._colorbar.vmax = value

        vmin_str, vmax_str = self._get_vmin_vmax_str()
        self._text_vmax.offset = (-120, self._linear_region_selector.selection[1], 0)
        self._text_vmax.text = vmax_str

    def set_data(self, data, reset_vmin_vmax: bool = True):
        hist, edges, hist_scaled, edges_flanked = self._calculate_histogram(data)

        line_data = np.column_stack([hist_scaled, edges_flanked])

        # set x and y vals
        self._histogram_line.data[:, :2] = line_data

        bounds = (edges[0], edges[-1])
        limits = (edges_flanked[0], edges_flanked[-11])
        origin = (hist_scaled.max() / 2, 0)

        if reset_vmin_vmax:
            # reset according to the new data
            self._linear_region_selector.limits = limits
            self._linear_region_selector.selection = bounds
        else:
            with pause_events(self._linear_region_selector, *self.images):
                # don't change the current selection
                self._linear_region_selector.limits = limits

        self._data = weakref.proxy(data)

        if self._colorbar is not None:
            self._colorbar.clear_event_handlers()
            self.world_object.remove(self._colorbar.world_object)

        if self.images[0].cmap is not None:
            self._colorbar: ImageGraphic = self._make_colorbar(edges_flanked)
            self._colorbar.add_event_handler(self._open_cmap_picker, "click")

            self.world_object.add(self._colorbar.world_object)
        else:
            self._colorbar = None
            self._cmap = None

        # reset plotarea dims
        self._plot_area.auto_scale()

    @property
    def images(self) -> tuple[ImageGraphic | ImageVolumeGraphic]:
        return self._images

    @images.setter
    def images(self, images):
        if isinstance(images, (ImageGraphic, ImageVolumeGraphic)):
            images = (images,)
        elif isinstance(images, Sequence):
            if not all(
                [isinstance(ig, (ImageGraphic, ImageVolumeGraphic)) for ig in images]
            ):
                raise TypeError(
                    f"`images` argument must be an ImageGraphic, ImageVolumeGraphic, or a "
                    f"tuple or list or ImageGraphic | ImageVolumeGraphic"
                )
        else:
            raise TypeError(
                f"`images` argument must be an ImageGraphic, ImageVolumeGraphic, or a "
                f"tuple or list or ImageGraphic | ImageVolumeGraphic"
            )

        if self._images is not None:
            for ig in self._images:
                # cleanup events from current image graphics
                ig_events = _get_image_graphic_events(ig)
                ig.remove_event_handler(self._image_cmap_handler, *ig_events)

        self._images = images

        ig_events = _get_image_graphic_events(self._images[0])

        for ig in self.images:
            ig.add_event_handler(self._image_cmap_handler, *ig_events)

    def _open_cmap_picker(self, ev):
        # check if right click
        if ev.button != 2:
            return

        pos = ev.x, ev.y

        self._plot_area.get_figure().open_popup("colormap-picker", pos, lut_tool=self)

    def _fpl_prepare_del(self):
        self._linear_region_selector._fpl_prepare_del()
        self._histogram_line._fpl_prepare_del()
        del self._histogram_line
        del self._linear_region_selector



import warnings
from numbers import Real
from typing import *
import numpy as np

import pygfx
from .._collection_base import GraphicCollection

from ...utils.enums import RenderQueue
from .._base import Graphic
from ..features import RectangleSelectionFeature
from ._base_selector import BaseSelector, MoveInfo


class RectangleSelector(BaseSelector):
    _features = {"selection": RectangleSelectionFeature}

    @property
    def parent(self) -> Graphic | None:
        """Graphic that selector is associated with."""
        return self._parent

    @property
    def selection(self) -> np.ndarray[float]:
        """
        (xmin, xmax, ymin, ymax) of the rectangle selection
        """
        return self._selection.value.copy()

    @selection.setter
    def selection(self, selection: Sequence[float]):
        # set (xmin, xmax, ymin, ymax) of the selector in data space
        graphic = self._parent

        if isinstance(graphic, GraphicCollection):
            pass

        self._selection.set_value(self, selection)

    @property
    def limits(self) -> Tuple[float, float, float, float]:
        """Return the limits of the selector."""
        return self._limits

    @limits.setter
    def limits(self, values: Tuple[float, float, float, float]):
        if len(values) != 4 or not all(map(lambda v: isinstance(v, Real), values)):
            raise TypeError("limits must be an iterable of two numeric values")
        self._limits = tuple(
            map(round, values)
        )  # if values are close to zero things get weird so round them
        self._selection._limits = self._limits

    def __init__(
        self,
        selection: Sequence[float],
        limits: Sequence[float],
        parent: Graphic = None,
        resizable: bool = True,
        fill_color=(0, 0, 0.35),
        edge_color=(0.8, 0.6, 0),
        edge_thickness: float = 8,
        vertex_color=(0.7, 0.4, 0),
        vertex_size: float = 8,
        arrow_keys_modifier: str = "Shift",
        name: str = None,
    ):
        """
        Create a RectangleSelector graphic which can be used to select a rectangular region of data.
        Allows sub-selecting data from a ``Graphic`` or from multiple Graphics.

        Parameters
        ----------
        selection: (float, float, float, float)
            the initial selection of the rectangle, ``(x_min, x_max, y_min, y_max)``

        limits: (float, float, float, float)
            limits of the selector, ``(x_min, x_max, y_min, y_max)``

        parent: Graphic, default ``None``
            associate this selector with a parent Graphic

        resizable: bool, default ``True``
            if ``True``, the edges can be dragged to resize the selection

        fill_color: str, array, or tuple
            fill color for the selector as a str or RGBA array

        edge_color: str, array, or tuple
            edge color for the selector as a str or RGBA array

        edge_thickness: float, default 8
            edge thickness

        vertex_color: str, array, or tuple
            vertex color for the selector as a str or RGBA array

        arrow_keys_modifier: str
            modifier key that must be pressed to initiate movement using arrow keys, must be one of:
            "Control", "Shift", "Alt" or ``None``

        name: str
            name for this selector graphic
        """

        if not len(selection) == 4 or not len(limits) == 4:
            raise ValueError()

        # lots of very close to zero values etc. so round them
        selection = tuple(map(round, selection))
        limits = tuple(map(round, limits))

        self._parent = parent
        self._limits = np.asarray(limits)
        self._resizable = resizable

        selection = np.asarray(selection)

        # world object for this will be a group
        # basic mesh for the fill area of the selector
        # line for each edge of the selector
        group = pygfx.Group()

        xmin, xmax, ymin, ymax = selection

        self._fill_color = pygfx.Color(fill_color)
        self._edge_color = pygfx.Color(edge_color)
        self._vertex_color = pygfx.Color(vertex_color)

        width = xmax - xmin
        height = ymax - ymin

        if width < 0 or height < 0:
            raise ValueError()

        self.fill = pygfx.Mesh(
            pygfx.box_geometry(width, height, 1),
            pygfx.MeshBasicMaterial(
                color=pygfx.Color(self.fill_color),
                alpha_mode="blend",
                opacity=0.4,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )

        self.fill.world.position = (0, 0, -2)

        group.add(self.fill)

        # position data for the left edge line
        left_line_data = np.array(
            [
                [xmin, ymin, 0],
                [xmin, ymax, 0],
            ]
        ).astype(np.float32)

        left_line = pygfx.Line(
            pygfx.Geometry(positions=left_line_data.copy()),
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        # position data for the right edge line
        right_line_data = np.array(
            [
                [xmax, ymin, 0],
                [xmax, ymax, 0],
            ]
        ).astype(np.float32)

        right_line = pygfx.Line(
            pygfx.Geometry(positions=right_line_data.copy()),
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        # position data for the left edge line
        bottom_line_data = np.array(
            [
                [xmin, ymax, 0],
                [xmax, ymax, 0],
            ]
        ).astype(np.float32)

        bottom_line = pygfx.Line(
            pygfx.Geometry(positions=bottom_line_data.copy()),
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        # position data for the right edge line
        top_line_data = np.array(
            [
                [xmin, ymin, 0],
                [xmax, ymin, 0],
            ]
        ).astype(np.float32)

        top_line = pygfx.Line(
            pygfx.Geometry(positions=top_line_data.copy()),
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        self.edges: Tuple[pygfx.Line, pygfx.Line, pygfx.Line, pygfx.Line] = (
            left_line,
            right_line,
            bottom_line,
            top_line,
        )  # left line, right line, bottom line, top line

        # add the edge lines

        for edge in self.edges:
            edge.render_order = 1
            group.add(*self.edges)

        # vertices
        top_left_vertex_data = (xmin, ymax, 1)
        top_right_vertex_data = (xmax, ymax, 1)
        bottom_left_vertex_data = (xmin, ymin, 1)
        bottom_right_vertex_data = (xmax, ymin, 1)

        top_left_vertex = pygfx.Points(
            pygfx.Geometry(positions=[top_left_vertex_data], sizes=[vertex_size]),
            pygfx.PointsMarkerMaterial(
                marker="square",
                size=vertex_size,
                color=self.vertex_color,
                size_mode="vertex",
                edge_color=self.vertex_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        top_right_vertex = pygfx.Points(
            pygfx.Geometry(positions=[top_right_vertex_data], sizes=[vertex_size]),
            pygfx.PointsMarkerMaterial(
                marker="square",
                size=vertex_size,
                color=self.vertex_color,
                size_mode="vertex",
                edge_color=self.vertex_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        bottom_left_vertex = pygfx.Points(
            pygfx.Geometry(positions=[bottom_left_vertex_data], sizes=[vertex_size]),
            pygfx.PointsMarkerMaterial(
                marker="square",
                size=vertex_size,
                color=self.vertex_color,
                size_mode="vertex",
                edge_color=self.vertex_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        bottom_right_vertex = pygfx.Points(
            pygfx.Geometry(positions=[bottom_right_vertex_data], sizes=[vertex_size]),
            pygfx.PointsMarkerMaterial(
                marker="square",
                size=vertex_size,
                color=self.vertex_color,
                size_mode="vertex",
                edge_color=self.vertex_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
            ),
        )

        self.vertices: Tuple[pygfx.Points, pygfx.Points, pygfx.Points, pygfx.Points] = (
            bottom_left_vertex,
            bottom_right_vertex,
            top_left_vertex,
            top_right_vertex,
        )

        for vertex in self.vertices:
            vertex.render_order = 2
            group.add(vertex)

        self._selection = RectangleSelectionFeature(selection, limits=self._limits)

        # include parent offset
        if parent is not None:
            offset = (parent.offset[0], parent.offset[1], 0)
        else:
            offset = (0, 0, 0)

        BaseSelector.__init__(
            self,
            edges=self.edges,
            fill=(self.fill,),
            vertices=self.vertices,
            hover_responsive=(*self.edges, *self.vertices),
            arrow_keys_modifier=arrow_keys_modifier,
            parent=parent,
            name=name,
            offset=offset,
        )

        self._set_world_object(group)

        self.selection = selection

    def get_selected_data(
        self, graphic: Graphic = None, mode: str = "full"
    ) -> Union[np.ndarray, List[np.ndarray]]:
        """
        Get the ``Graphic`` data bounded by the current selection.
        Returns a view of the data array.

        If the ``Graphic`` is a collection, such as a ``LineStack``, it returns a list of views of the full array.
        Can be performed on the ``parent`` Graphic or on another graphic by passing to the ``graphic`` arg.

        Parameters
        ----------
        graphic: Graphic, optional, default ``None``
            if provided, returns the data selection from this graphic instead of the graphic set as ``parent``
        mode: str, default 'full'
            One of 'full', 'partial', or 'ignore'. Indicates how selected data should be returned based on the
            selectors position over the graphic. Only used for ``LineGraphic``, ``LineCollection``, and ``LineStack``
            | If 'full', will return all data bounded by the x and y limits of the selector even if partial indices
            along one axis are not fully covered by the selector.
            | If 'partial' will return only the data that is bounded by the selector, missing indices not bounded by the
            selector will be set to NaNs
            | If 'ignore', will only return data for graphics that have indices completely bounded by the selector

        Returns
        -------
        np.ndarray or List[np.ndarray]
            view or list of views of the full array, returns empty array if selection is empty
        """
        source = self._get_source(graphic)
        ixs = self.get_selected_indices(source)

        # do not need to check for mode for images, because the selector is bounded by the image shape
        # will always be `full`
        if "Image" in source.__class__.__name__:
            row_ixs, col_ixs = ixs
            row_slice = slice(row_ixs[0], row_ixs[-1] + 1)
            col_slice = slice(col_ixs[0], col_ixs[-1] + 1)

            return source.data[row_slice, col_slice]

        if mode not in ["full", "partial", "ignore"]:
            raise ValueError(
                f"`mode` must be one of 'full', 'partial', or 'ignore', you have passed {mode}"
            )
        if "Line" in source.__class__.__name__:
            if isinstance(source, GraphicCollection):
                data_selections: List[np.ndarray] = list()

                for i, g in enumerate(source.graphics):
                    # want to keep same length as the original line collection
                    if ixs[i].size == 0:
                        data_selections.append(
                            np.array([], dtype=np.float32).reshape(0, 3)
                        )
                    else:
                        # s gives entire slice of data along the x
                        s = slice(
                            ixs[i][0], ixs[i][-1] + 1
                        )  # add 1 because these are direct indices
                        # slices n_datapoints dim

                        # calculate missing ixs using set difference
                        # then calculate shift
                        missing_ixs = (
                            np.setdiff1d(np.arange(ixs[i][0], ixs[i][-1] + 1), ixs[i])
                            - ixs[i][0]
                        )

                        match mode:
                            # take all ixs, ignore missing
                            case "full":
                                data_selections.append(g.data[s])
                            # set missing ixs data to NaNs
                            case "partial":
                                if len(missing_ixs) > 0:
                                    data = g.data[s].copy()
                                    data[missing_ixs] = np.nan
                                    data_selections.append(data)
                                else:
                                    data_selections.append(g.data[s])
                            # ignore lines that do not have full ixs to start
                            case "ignore":
                                if len(missing_ixs) > 0:
                                    data_selections.append(
                                        np.array([], dtype=np.float32).reshape(0, 3)
                                    )
                                else:
                                    data_selections.append(g.data[s])
                return data_selections
            else:  # for lines
                if ixs.size == 0:
                    # empty selection
                    return np.array([], dtype=np.float32).reshape(0, 3)

                s = slice(
                    ixs[0], ixs[-1] + 1
                )  # add 1 to end because these are direct indices
                # slices n_datapoints dim
                # slice with min, max is faster than using all the indices

                # get missing ixs
                missing_ixs = np.setdiff1d(np.arange(ixs[0], ixs[-1] + 1), ixs) - ixs[0]

                match mode:
                    # return all, do not care about missing
                    case "full":
                        return source.data[s]
                    # set missing to NaNs
                    case "partial":
                        if len(missing_ixs) > 0:
                            data = source.data[s].copy()
                            data[missing_ixs] = np.nan
                            return data
                        else:
                            return source.data[s]
                    # missing means nothing will be returned even if selector is partially over data
                    # warn the user and return empty
                    case "ignore":
                        if len(missing_ixs) > 0:
                            warnings.warn(
                                "You have selected 'ignore' mode. Selected graphic has incomplete indices. "
                                "Move the selector or change the mode to one of `partial` or `full`."
                            )
                            return np.array([], dtype=np.float32)
                        else:
                            return source.data[s]

    def get_selected_indices(
        self, graphic: Graphic = None
    ) -> np.ndarray | tuple[np.ndarray]:
        """
        Returns the indices of the ``Graphic`` data bounded by the current selection.

        These are the data indices which correspond to the data under the selector.

        Parameters
        ----------
        graphic: Graphic, default ``None``
            If provided, returns the selection indices from this graphic instead of the graphic set as ``parent``

        Returns
        -------
        Union[np.ndarray, List[np.ndarray]]
            data indicies of the selection
            | tuple of [row_indices, col_indices] if the graphic is an image
            | list of indices along the x-dimension for each line if graphic is a line collection
            | array of indices along the x-dimension if graphic is a line
        """
        # get indices from source
        source = self._get_source(graphic)

        # selector (xmin, xmax, ymin, ymax) values
        xmin, xmax, ymin, ymax = self.selection

        # image data does not need to check for mode because the selector is always bounded
        # to the image
        if "Image" in source.__class__.__name__:
            col_ixs = np.arange(xmin, xmax, dtype=int)
            row_ixs = np.arange(ymin, ymax, dtype=int)
            return row_ixs, col_ixs

        if "Line" in source.__class__.__name__:
            if isinstance(source, GraphicCollection):
                ixs = list()
                for g in source.graphics:
                    data = g.data.value
                    g_ixs = np.where(
                        (data[:, 0] >= xmin - g.offset[0])
                        & (data[:, 0] <= xmax - g.offset[0])
                        & (data[:, 1] >= ymin - g.offset[1])
                        & (data[:, 1] <= ymax - g.offset[1])
                    )[0]
                    ixs.append(g_ixs)
            else:
                # map only this graphic
                data = source.data.value
                ixs = np.where(
                    (data[:, 0] >= xmin)
                    & (data[:, 0] <= xmax)
                    & (data[:, 1] >= ymin)
                    & (data[:, 1] <= ymax)
                )[0]

            return ixs

    def _move_graphic(self, move_info: MoveInfo):
        # If this the first move in this drag, store initial selection
        if move_info.start_selection is None:
            move_info.start_selection = self.selection

        # add delta to current min, max to get new positions
        deltax, deltay = move_info.delta[0], move_info.delta[1]

        # Get original selection
        xmin, xmax, ymin, ymax = move_info.start_selection

        # move entire selector if source is fill
        if move_info.source == self.fill:
            # Limit the delta to avoid weird resizine behavior
            min_deltax = self.limits[0] - xmin
            max_deltax = self.limits[1] - xmax
            min_deltay = self.limits[2] - ymin
            max_deltay = self.limits[3] - ymax
            deltax = np.clip(deltax, min_deltax, max_deltax)
            deltay = np.clip(deltay, min_deltay, max_deltay)
            # Update all bounds with equal amount
            self._selection.set_value(
                self, (xmin + deltax, xmax + deltax, ymin + deltay, ymax + deltay)
            )
            return

        # if selector not resizable return
        if not self._resizable:
            return

        xmin_new = min(xmin + deltax, xmax)
        xmax_new = max(xmax + deltax, xmin)
        ymin_new = min(ymin + deltay, ymax)
        ymax_new = max(ymax + deltay, ymin)

        if move_info.source == self.vertices[0]:  # bottom left
            self._selection.set_value(self, (xmin_new, xmax, ymin_new, ymax))
        if move_info.source == self.vertices[1]:  # bottom right
            self._selection.set_value(self, (xmin, xmax_new, ymin_new, ymax))
        if move_info.source == self.vertices[2]:  # top left
            self._selection.set_value(self, (xmin_new, xmax, ymin, ymax_new))
        if move_info.source == self.vertices[3]:  # top right
            self._selection.set_value(self, (xmin, xmax_new, ymin, ymax_new))
        # if event source was an edge and selector is resizable, move the edge that caused the event
        if move_info.source == self.edges[0]:
            self._selection.set_value(self, (xmin_new, xmax, ymin, ymax))
        if move_info.source == self.edges[1]:
            self._selection.set_value(self, (xmin, xmax_new, ymin, ymax))
        if move_info.source == self.edges[2]:
            self._selection.set_value(self, (xmin, xmax, ymin_new, ymax))
        if move_info.source == self.edges[3]:
            self._selection.set_value(self, (xmin, xmax, ymin, ymax_new))

    def _move_to_pointer(self, ev):
        pass



import importlib
from itertools import product

import numpy as np

import pygfx
from pygfx import WgpuRenderer, Texture, Renderer

from ..utils.gui import BaseRenderCanvas, RenderCanvas

try:
    import imgui_bundle
except ImportError:
    IMGUI = False
else:
    IMGUI = True


# number of pixels taken by the imgui toolbar when present
IMGUI_TOOLBAR_HEIGHT = 36


def make_canvas_and_renderer(
    canvas: str | BaseRenderCanvas | Texture | None,
    renderer: Renderer | None,
    canvas_kwargs: dict,
):
    """
    Parses arguments and returns the appropriate canvas and renderer instances
    as a tuple (canvas, renderer)
    """

    if canvas is None:
        canvas = RenderCanvas(**canvas_kwargs)
    elif isinstance(canvas, str):
        import rendercanvas

        m = importlib.import_module("rendercanvas." + canvas)
        canvas = m.RenderCanvas(**canvas_kwargs)
    elif not isinstance(canvas, (BaseRenderCanvas, Texture)):
        raise TypeError(
            f"canvas option must either be a valid BaseRenderCanvas implementation, a pygfx Texture"
            f" or a str with the gui backend name, valid str are: 'qt', 'glfw', 'jupyter', 'wx', and 'offscreen'"
        )

    if renderer is None:
        renderer = WgpuRenderer(canvas)
    elif not isinstance(renderer, Renderer):
        raise TypeError(
            f"renderer option must be a pygfx.Renderer instance such as pygfx.WgpuRenderer"
        )

    return canvas, renderer


def create_camera(
    camera_type: pygfx.PerspectiveCamera | str,
) -> pygfx.PerspectiveCamera:
    if isinstance(camera_type, pygfx.PerspectiveCamera):
        return camera_type

    elif camera_type == "2d":
        # use perspective for orthographic, makes it easier to then switch to controllers that make sense with fov > 0
        return pygfx.PerspectiveCamera(fov=0)

    elif camera_type == "3d":
        return pygfx.PerspectiveCamera()

    else:
        raise ValueError(
            "camera must be one of: '2d', '3d' or an instance of pygfx.PerspectiveCamera"
        )


controller_types = {
    "fly": pygfx.FlyController,
    "panzoom": pygfx.PanZoomController,
    "trackball": pygfx.TrackballController,
    "orbit": pygfx.OrbitController,
}


def create_controller(
    controller_type: pygfx.Controller | None | str,
    camera: pygfx.PerspectiveCamera,
) -> pygfx.Controller:
    """
    Creates the controllers and adds the camera to it.
    """
    if isinstance(controller_type, pygfx.Controller):
        controller_type.add_camera(camera)
        return controller_type

    if controller_type is None:
        # default controllers
        if camera.fov == 0:
            # default for orthographic
            return pygfx.PanZoomController(camera)
        else:
            return pygfx.FlyController(camera)

    if controller_type not in controller_types.keys():
        raise KeyError(
            f"controller must be a valid pygfx.Controller or one of: "
            f"{list(controller_types.keys())}, you have passed: {controller_type}"
        )

    return controller_types[controller_type](camera)


def get_extents_from_grid(
    shape: tuple[int, int],
) -> list[tuple[float, float, float, float]]:
    """create fractional extents from a given grid shape"""
    x_min = np.arange(0, 1, (1 / shape[1]))
    x_max = x_min + 1 / shape[1]
    y_min = np.arange(0, 1, (1 / shape[0]))
    y_max = y_min + 1 / shape[0]

    extents = list()
    for row_ix, col_ix in product(range(shape[0]), range(shape[1])):
        extent = x_min[col_ix], x_max[col_ix], y_min[row_ix], y_max[row_ix]
        extents.append(extent)

    return extents



from typing import Sequence

import numpy as np

from ..graphics._base import Graphic
from ..graphics._collection_base import GraphicCollection


def get_nearest_graphics_indices(
    pos: tuple[float, float] | tuple[float, float, float],
    graphics: Sequence[Graphic] | GraphicCollection,
) -> np.ndarray[int]:
    """
    Returns indices of the nearest ``graphics`` to the passed position ``pos`` in world space
    in order of closest to furtherst. Uses the distance between ``pos`` and the center of the
    bounding sphere for each graphic.

    Parameters
    ----------
    pos: (x, y) | (x, y, z)
        position in world space, z-axis is ignored when calculating L2 norms if ``pos`` is 2D

    graphics: Sequence, i.e. array, list, tuple, etc. of Graphic | GraphicCollection
        the graphics from which to return a sorted array of graphics in order of closest
        to furthest graphic

    Returns
    -------
    ndarray[int]
        indices of the nearest nearest graphics to ``pos`` in order

    """
    if isinstance(graphics, GraphicCollection):
        graphics = graphics.graphics

    if not all(isinstance(g, Graphic) for g in graphics):
        raise TypeError("all elements of `graphics` must be Graphic objects")

    pos = np.asarray(pos).ravel()

    if pos.shape != (2,) and pos.shape != (3,):
        raise TypeError(
            f"pos.shape must be (2,) or (3,), the shape of pos you have passed is: {pos.shape}"
        )

    # get centers
    centers = np.empty(shape=(len(graphics), len(pos)))
    for i in range(centers.shape[0]):
        centers[i] = graphics[i].world_object.get_world_bounding_sphere()[: len(pos)]

    # l2
    distances = np.linalg.norm(centers[:, : len(pos)] - pos, ord=2, axis=1)

    sort_indices = np.argsort(distances)
    return sort_indices


def get_nearest_graphics(
    pos: tuple[float, float] | tuple[float, float, float],
    graphics: Sequence[Graphic] | GraphicCollection,
) -> np.ndarray[Graphic]:
    """
    Returns the nearest ``graphics`` to the passed position ``pos`` in world space.
    Uses the distance between ``pos`` and the center of the bounding sphere for each graphic.

    Parameters
    ----------
    pos: (x, y) | (x, y, z)
        position in world space, z-axis is ignored when calculating L2 norms if ``pos`` is 2D

    graphics: Sequence, i.e. array, list, tuple, etc. of Graphic | GraphicCollection
        the graphics from which to return a sorted array of graphics in order of closest
        to furthest graphic

    Returns
    -------
    ndarray[Graphic]
        nearest graphics to ``pos`` in order

    """
    sort_indices = get_nearest_graphics_indices(pos, graphics)
    return np.asarray(graphics)[sort_indices]



import wgpu
from pygfx.renderers.wgpu import select_adapter as pygfx_select_adapter
from pygfx import print_wgpu_report as pygfx_print_wgpu_report


def enumerate_adapters() -> list[wgpu.GPUAdapter]:
    return wgpu.gpu.enumerate_adapters_sync()


enumerate_adapters.__doc__ = wgpu.gpu.enumerate_adapters_async.__doc__


def select_adapter(adapter: wgpu.GPUAdapter):
    return pygfx_select_adapter(adapter)


select_adapter.__doc__ = pygfx_select_adapter.__doc__


def print_wgpu_report():
    return pygfx_print_wgpu_report()


print_wgpu_report.__doc__ = pygfx_print_wgpu_report.__doc__



from typing import Any

import numpy as np

import pygfx
from ._base import Graphic
from .features import (
    VertexPositions,
    VertexColors,
    UniformColor,
    VertexCmap,
    PointsSizesFeature,
    SizeSpace,
)


class PositionsGraphic(Graphic):
    """Base class for LineGraphic and ScatterGraphic"""

    @property
    def data(self) -> VertexPositions:
        """Get or set the graphic's data"""
        return self._data

    @data.setter
    def data(self, value):
        self._data[:] = value

    @property
    def colors(self) -> VertexColors | pygfx.Color:
        """Get or set the colors"""
        if isinstance(self._colors, VertexColors):
            return self._colors

        elif isinstance(self._colors, UniformColor):
            return self._colors.value

    @colors.setter
    def colors(self, value: str | np.ndarray | tuple[float] | list[float] | list[str]):
        if isinstance(self._colors, VertexColors):
            self._colors[:] = value

        elif isinstance(self._colors, UniformColor):
            self._colors.set_value(self, value)

    @property
    def cmap(self) -> VertexCmap:
        """
        Control the cmap or cmap transform

        For supported colormaps see the ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/
        """
        return self._cmap

    @cmap.setter
    def cmap(self, name: str):
        if self._cmap is None:
            raise BufferError("Cannot use cmap with uniform_colors=True")

        self._cmap[:] = name

    @property
    def size_space(self):
        """
        The coordinate space in which the size is expressed ('screen', 'world', 'model')

        See https://docs.pygfx.org/stable/_autosummary/utils/utils/enums/pygfx.utils.enums.CoordSpace.html#pygfx.utils.enums.CoordSpace for available options.
        """
        return self._size_space.value

    @size_space.setter
    def size_space(self, value: str):
        self._size_space.set_value(self, value)

    def __init__(
        self,
        data: Any,
        colors: str | np.ndarray | tuple[float] | list[float] | list[str] = "w",
        uniform_color: bool = False,
        cmap: str | VertexCmap = None,
        cmap_transform: np.ndarray = None,
        isolated_buffer: bool = True,
        size_space: str = "screen",
        *args,
        **kwargs,
    ):
        if isinstance(data, VertexPositions):
            self._data = data
        else:
            self._data = VertexPositions(data, isolated_buffer=isolated_buffer)

        if cmap_transform is not None and cmap is None:
            raise ValueError("must pass `cmap` if passing `cmap_transform`")

        if cmap is not None:
            # if a cmap is specified it overrides colors argument
            if uniform_color:
                raise TypeError("Cannot use cmap if uniform_color=True")

            if isinstance(cmap, str):
                # make colors from cmap
                if isinstance(colors, VertexColors):
                    # share buffer with existing colors instance for the cmap
                    self._colors = colors
                    self._colors._shared += 1
                else:
                    # create vertex colors buffer
                    self._colors = VertexColors("w", n_colors=self._data.value.shape[0])
                    # make cmap using vertex colors buffer
                    self._cmap = VertexCmap(
                        self._colors,
                        cmap_name=cmap,
                        transform=cmap_transform,
                    )
            elif isinstance(cmap, VertexCmap):
                # use existing cmap instance
                self._cmap = cmap
                self._colors = cmap._vertex_colors
            else:
                raise TypeError(
                    "`cmap` argument must be a <str> cmap name or an existing `VertexCmap` instance"
                )
        else:
            # no cmap given
            if isinstance(colors, VertexColors):
                # share buffer with existing colors instance
                self._colors = colors
                self._colors._shared += 1
                # blank colormap instance
                self._cmap = VertexCmap(self._colors, cmap_name=None, transform=None)
            else:
                if uniform_color:
                    if not isinstance(colors, str):  # not a single color
                        if not len(colors) in [3, 4]:  # not an RGB(A) array
                            raise TypeError(
                                "must pass a single color if using `uniform_colors=True`"
                            )
                    self._colors = UniformColor(colors)
                    self._cmap = None
                else:
                    self._colors = VertexColors(
                        colors, n_colors=self._data.value.shape[0]
                    )
                    self._cmap = VertexCmap(
                        self._colors, cmap_name=None, transform=None
                    )

        self._size_space = SizeSpace(size_space)
        super().__init__(*args, **kwargs)



from copy import deepcopy
from typing import Callable
from warnings import warn

import numpy as np

from rendercanvas import BaseRenderCanvas

from ...layouts import ImguiFigure as Figure
from ...graphics import ImageGraphic
from ...utils import calculate_figure_shape, quick_min_max
from ...tools import HistogramLUTTool
from ._sliders import ImageWidgetSliders


# Number of dimensions that represent one image/one frame
# For grayscale shape will be [n_rows, n_cols], i.e. 2 dims
# For RGB(A) shape will be [n_rows, n_cols, c] where c is of size 3 (RGB) or 4 (RGBA)
IMAGE_DIM_COUNTS = {"gray": 2, "rgb": 3}

# Map boolean (indicating whether we use RGB or grayscale) to the string. Used to index RGB_DIM_MAP
RGB_BOOL_MAP = {False: "gray", True: "rgb"}

# Dimensions that can be scrolled from a given data array
SCROLLABLE_DIMS_ORDER = {
    0: "",
    1: "t",
    2: "tz",
}

ALLOWED_SLIDER_DIMS = {0: "t", 1: "z"}

ALLOWED_WINDOW_DIMS = {"t", "z"}


def _is_arraylike(obj) -> bool:
    """
    Checks if the object is array-like.
    For now just checks if obj has `__getitem__()`
    """
    for attr in ["__getitem__", "shape", "ndim"]:
        if not hasattr(obj, attr):
            return False

    return True


class _WindowFunctions:
    """Stores window function and window size"""

    def __init__(self, image_widget, func: callable, window_size: int):
        self._image_widget = image_widget
        self._func = None
        self.func = func

        self._window_size = 0
        self.window_size = window_size

    @property
    def func(self) -> callable:
        """Get or set the function"""
        return self._func

    @func.setter
    def func(self, func: callable):
        self._func = func

        # force update
        self._image_widget.current_index = self._image_widget.current_index

    @property
    def window_size(self) -> int:
        """Get or set window size"""
        return self._window_size

    @window_size.setter
    def window_size(self, ws: int):
        if ws is None:
            self._window_size = None
            return

        if not isinstance(ws, int):
            raise TypeError("window size must be an int")

        if ws < 3:
            warn(
                f"Invalid 'window size' value for function: {self.func}, "
                f"setting 'window size' = None for this function. "
                f"Valid values are integers >= 3."
            )
            self.window_size = None
            return

        if ws % 2 == 0:
            ws += 1

        self._window_size = ws

        self._image_widget.current_index = self._image_widget.current_index

    def __repr__(self):
        return f"func: {self.func}, window_size: {self.window_size}"


class ImageWidget:
    @property
    def figure(self) -> Figure:
        """
        ``Figure`` used by `ImageWidget`.
        """
        return self._figure

    @property
    def managed_graphics(self) -> list[ImageGraphic]:
        """List of ``ImageWidget`` managed graphics."""
        iw_managed = list()
        for subplot in self.figure:
            # empty subplots will not have any image widget data
            if len(subplot.graphics) > 0:
                iw_managed.append(subplot["image_widget_managed"])
        return iw_managed

    @property
    def cmap(self) -> list[str]:
        cmaps = list()
        for g in self.managed_graphics:
            cmaps.append(g.cmap)

        return cmaps

    @cmap.setter
    def cmap(self, names: str | list[str]):
        if isinstance(names, list):
            if not all([isinstance(n, str) for n in names]):
                raise TypeError(
                    f"Must pass cmap name as a `str` of list of `str`, you have passed:\n{names}"
                )

            if not len(names) == len(self.managed_graphics):
                raise IndexError(
                    f"If passing a list of cmap names, the length of the list must be the same as the number of "
                    f"image widget subplots. You have passed: {len(names)} cmap names and have "
                    f"{len(self.managed_graphics)} image widget subplots"
                )

            for name, g in zip(names, self.managed_graphics):
                g.cmap = name

        elif isinstance(names, str):
            for g in self.managed_graphics:
                g.cmap = names

    @property
    def data(self) -> list[np.ndarray]:
        """data currently displayed in the widget"""
        return self._data

    @property
    def ndim(self) -> int:
        """Number of dimensions of grayscale data displayed in the widget (it will be 1 more for RGB(A) data)"""
        return self._ndim

    @property
    def n_scrollable_dims(self) -> list[int]:
        """
        list indicating the number of dimenensions that are scrollable for each data array
        All other dimensions are frame/image data, i.e. [rows, cols] or [rows, cols, rgb(a)]
        """
        return self._n_scrollable_dims

    @property
    def slider_dims(self) -> list[str]:
        """the dimensions that the sliders index"""
        return self._slider_dims

    @property
    def current_index(self) -> dict[str, int]:
        """
        Get or set the current index

        Returns
        -------
        index: Dict[str, int]
            | ``dict`` for indexing each dimension, provide a ``dict`` with indices for all dimensions used by sliders
            or only a subset of dimensions used by the sliders.
            | example: if you have sliders for dims "t" and "z", you can pass either ``{"t": 10}`` to index to position
            10 on dimension "t" or ``{"t": 5, "z": 20}`` to index to position 5 on dimension "t" and position 20 on
            dimension "z" simultaneously.

        """
        return self._current_index

    @current_index.setter
    def current_index(self, index: dict[str, int]):
        if not self._initialized:
            return

        if self._reentrant_block:
            return

        try:
            self._reentrant_block = True  # block re-execution until current_index has *fully* completed execution
            if not set(index.keys()).issubset(set(self._current_index.keys())):
                raise KeyError(
                    f"All dimension keys for setting `current_index` must be present in the widget sliders. "
                    f"The dimensions currently used for sliders are: {list(self.current_index.keys())}"
                )

            for k, val in index.items():
                if not isinstance(val, int):
                    raise TypeError("Indices for all dimensions must be int")
                if val < 0:
                    raise IndexError(
                        "negative indexing is not supported for ImageWidget"
                    )
                if val > self._dims_max_bounds[k]:
                    raise IndexError(
                        f"index {val} is out of bounds for dimension '{k}' "
                        f"which has a max bound of: {self._dims_max_bounds[k]}"
                    )

            self._current_index.update(index)

            for i, (ig, data) in enumerate(zip(self.managed_graphics, self.data)):
                frame = self._process_indices(data, self._current_index)
                frame = self._process_frame_apply(frame, i)
                ig.data = frame

            # call any event handlers
            for handler in self._current_index_changed_handlers:
                handler(self.current_index)
        except Exception as exc:
            # raise original exception
            raise exc  # current_index setter has raised. The lines above below are probably more relevant!
        finally:
            # set_value has finished executing, now allow future executions
            self._reentrant_block = False

    @property
    def n_img_dims(self) -> list[int]:
        """
        list indicating the number of dimensions that contain image/single frame data for each data array.
        if 2: data are grayscale, i.e. [x, y] dims, if 3: data are [x, y, c] where c is RGB or RGBA,
        this is the complement of `n_scrollable_dims`
        """
        return self._n_img_dims

    def _get_n_scrollable_dims(self, curr_arr: np.ndarray, rgb: bool) -> list[int]:
        """
        For a given ``array`` displayed in the ImageWidget, this function infers how many of the dimensions are
        supported by sliders (aka scrollable). Ex: "xy" data has 0 scrollable dims, "txy" has 1, "tzxy" has 2.

        Parameters
        ----------
        curr_arr: np.ndarray
            np.ndarray or a list of array-like

        rgb: bool
            True if we view this as RGB(A) and False if grayscale

        Returns
        -------
        int
            Number of scrollable dimensions for each ``array`` in the dataset.
        """

        n_img_dims = IMAGE_DIM_COUNTS[RGB_BOOL_MAP[rgb]]
        # Make sure each image stack at least ``n_img_dims`` dimensions
        if len(curr_arr.shape) < n_img_dims:
            raise ValueError(
                f"Your array has shape {curr_arr.shape} "
                f"but you specified that each image in your array is {n_img_dims}D "
            )

        # If RGB(A), last dim must be 3 or 4
        if n_img_dims == 3:
            if not (curr_arr.shape[-1] == 3 or curr_arr.shape[-1] == 4):
                raise ValueError(
                    f"Expected size 3 or 4 for last dimension of RGB(A) array, got: {curr_arr.shape[-1]}."
                )

        n_scrollable_dims = len(curr_arr.shape) - n_img_dims

        if n_scrollable_dims not in SCROLLABLE_DIMS_ORDER.keys():
            raise ValueError(f"Array had shape {curr_arr.shape} which is not supported")

        return n_scrollable_dims

    def __init__(
        self,
        data: np.ndarray | list[np.ndarray],
        window_funcs: dict[str, tuple[Callable, int]] = None,
        frame_apply: Callable | dict[int, Callable] = None,
        figure_shape: tuple[int, int] = None,
        names: list[str] = None,
        figure_kwargs: dict = None,
        histogram_widget: bool = True,
        rgb: bool | list[bool] = None,
        cmap: str = "plasma",
        graphic_kwargs: dict = None,
    ):
        """
        This widget facilitates high-level navigation through image stacks, which are arrays containing one or more
        images. It includes sliders for key dimensions such as "t" (time) and "z", enabling users to smoothly navigate
        through one or multiple image stacks simultaneously.

        Allowed dimensions orders for each image stack: Note that each has a an optional (c) channel which refers to
        RGB(A) a channel. So this channel should be either 3 or 4.

        ======= ==========
        n_dims  dims order
        ======= ==========
        2       "xy(c)"
        3       "txy(c)"
        4       "tzxy(c)"
        ======= ==========

        Parameters
        ----------
        data: Union[np.ndarray, List[np.ndarray]
            array-like or a list of array-like

        window_funcs: dict[str, tuple[Callable, int]], i.e. {"t" or "z": (callable, int)}
            | Apply function(s) with rolling windows along "t" and/or "z" dimensions of the `data` arrays.
            | Pass a dict in the form: {dimension: (func, window_size)}, `func` must take a slice of the data array as
            | the first argument and must take `axis` as a kwarg.
            | Ex: mean along "t" dimension: {"t": (np.mean, 11)}, if `current_index` of "t" is 50, it will pass frames
            | 45 to 55 to `np.mean` with `axis=0`.
            | Ex: max along z dim: {"z": (np.max, 3)}, passes current, previous & next frame to `np.max` with `axis=1`

        frame_apply: Union[callable, Dict[int, callable]]
            | Apply function(s) to `data` arrays before to generate final 2D image that is displayed.
            | Ex: apply a spatial gaussian filter
            | Pass a single function or a dict of functions to apply to each array individually
            | examples: ``{array_index: to_grayscale}``, ``{0: to_grayscale, 2: threshold_img}``
            | "array_index" is the position of the corresponding array in the data list.
            | if `window_funcs` is used, then this function is applied after `window_funcs`
            | this function must be a callable that returns a 2D array
            | example use case: converting an RGB frame from video to a 2D grayscale frame

        figure_shape: Optional[Tuple[int, int]]
            manually provide the shape for the Figure, otherwise the number of rows and columns is estimated

        figure_kwargs: dict, optional
            passed to ``Figure``

        names: Optional[str]
            gives names to the subplots

        histogram_widget: bool, default False
            make histogram LUT widget for each subplot

        rgb: bool | list[bool], default None
            bool or list of bool for each input data array in the ImageWidget, indicating whether the corresponding
            data arrays are grayscale or RGB(A).

        graphic_kwargs: Any
            passed to each ImageGraphic in the ImageWidget figure subplots

        """
        self._initialized = False

        if figure_kwargs is None:
            figure_kwargs = dict()

        if _is_arraylike(data):
            data = [data]

        if isinstance(data, list):
            # verify that it's a list of np.ndarray
            if all([_is_arraylike(d) for d in data]):
                # Grid computations
                if figure_shape is None:
                    if "shape" in figure_kwargs:
                        figure_shape = figure_kwargs["shape"]
                    else:
                        figure_shape = calculate_figure_shape(len(data))

                # Regardless of how figure_shape is computed, below code
                # verifies that figure shape is large enough for the number of image arrays passed
                if figure_shape[0] * figure_shape[1] < len(data):
                    original_shape = (figure_shape[0], figure_shape[1])
                    figure_shape = calculate_figure_shape(len(data))
                    warn(
                        f"Original `figure_shape` was: {original_shape} "
                        f" but data length is {len(data)}"
                        f" Resetting figure shape to: {figure_shape}"
                    )

                self._data: list[np.ndarray] = data

                # Establish number of image dimensions and number of scrollable dimensions for each array
                if rgb is None:
                    rgb = [False] * len(self.data)
                if isinstance(rgb, bool):
                    rgb = [rgb] * len(self.data)
                if not isinstance(rgb, list):
                    raise TypeError(
                        f"`rgb` parameter must be a bool or list of bool, a <{type(rgb)}> was provided"
                    )
                if not len(rgb) == len(self.data):
                    raise ValueError(
                        f"len(rgb) != len(data), {len(rgb)} != {len(self.data)}. These must be equal"
                    )

                self._rgb = rgb

                self._n_img_dims = [
                    IMAGE_DIM_COUNTS[RGB_BOOL_MAP[self._rgb[i]]]
                    for i in range(len(self.data))
                ]

                self._n_scrollable_dims = [
                    self._get_n_scrollable_dims(self.data[i], self._rgb[i])
                    for i in range(len(self.data))
                ]

                # Define ndim of ImageWidget instance as largest number of scrollable dims + 2 (grayscale dimensions)
                self._ndim = (
                    max(
                        [
                            self.n_scrollable_dims[i]
                            for i in range(len(self.n_scrollable_dims))
                        ]
                    )
                    + IMAGE_DIM_COUNTS[RGB_BOOL_MAP[False]]
                )

                if names is not None:
                    if not all([isinstance(n, str) for n in names]):
                        raise TypeError(
                            "optional argument `names` must be a list of str"
                        )

                    if len(names) != len(self.data):
                        raise ValueError(
                            "number of `names` for subplots must be same as the number of data arrays"
                        )

            else:
                raise TypeError(
                    f"If passing a list to `data` all elements must be an "
                    f"array-like type representing an n-dimensional image. "
                    f"You have passed the following types:\n"
                    f"{[type(a) for a in data]}"
                )
        else:
            raise TypeError(
                f"`data` must be an array-like type or a list of array-like."
                f"You have passed the following type {type(data)}"
            )

        # Sliders are made for all dimensions except the image dimensions
        self._slider_dims = list()
        max_scrollable = max(
            [self.n_scrollable_dims[i] for i in range(len(self.n_scrollable_dims))]
        )
        for dim in range(max_scrollable):
            if dim in ALLOWED_SLIDER_DIMS.keys():
                self.slider_dims.append(ALLOWED_SLIDER_DIMS[dim])

        self._frame_apply: dict[int, callable] = dict()

        if frame_apply is not None:
            if callable(frame_apply):
                self._frame_apply = frame_apply

            elif isinstance(frame_apply, dict):
                self._frame_apply: dict[int, callable] = dict.fromkeys(
                    list(range(len(self.data)))
                )

                # dict of {array: dims_order_str}
                for data_ix in list(frame_apply.keys()):
                    if not isinstance(data_ix, int):
                        raise TypeError("`frame_apply` dict keys must be <int>")
                    try:
                        self._frame_apply[data_ix] = frame_apply[data_ix]
                    except Exception:
                        raise IndexError(
                            f"key index {data_ix} out of bounds for `frame_apply`, the bounds are 0 - {len(self.data)}"
                        )
            else:
                raise TypeError(
                    f"`frame_apply` must be a callable or <Dict[int: callable]>, "
                    f"you have passed a: <{type(frame_apply)}>"
                )

        # current_index stores {dimension_index: slice_index} for every dimension
        self._current_index: dict[str, int] = {sax: 0 for sax in self.slider_dims}

        self._window_funcs = None
        self.window_funcs = window_funcs

        # get max bound for all data arrays for all slider dimensions and ensure compatibility across slider dims
        self._dims_max_bounds: dict[str, int] = {k: 0 for k in self.slider_dims}
        for i, _dim in enumerate(list(self._dims_max_bounds.keys())):
            for array, partition in zip(self.data, self.n_scrollable_dims):
                if partition <= i:
                    continue
                else:
                    if 0 < self._dims_max_bounds[_dim] != array.shape[i]:
                        raise ValueError(f"Two arrays differ along dimension {_dim}")
                    else:
                        self._dims_max_bounds[_dim] = max(
                            self._dims_max_bounds[_dim], array.shape[i]
                        )

        figure_kwargs_default = {"controller_ids": "sync", "names": names}

        # update the default kwargs with any user-specified kwargs
        # user specified kwargs will overwrite the defaults
        figure_kwargs_default.update(figure_kwargs)
        figure_kwargs_default["shape"] = figure_shape

        if graphic_kwargs is None:
            graphic_kwargs = dict()

        graphic_kwargs.update({"cmap": cmap})

        vmin_specified, vmax_specified = None, None
        if "vmin" in graphic_kwargs.keys():
            vmin_specified = graphic_kwargs.pop("vmin")
        if "vmax" in graphic_kwargs.keys():
            vmax_specified = graphic_kwargs.pop("vmax")

        self._figure: Figure = Figure(**figure_kwargs_default)

        self._histogram_widget = histogram_widget
        for data_ix, (d, subplot) in enumerate(zip(self.data, self.figure)):

            frame = self._process_indices(d, slice_indices=self._current_index)
            frame = self._process_frame_apply(frame, data_ix)

            if (vmin_specified is None) or (vmax_specified is None):
                # if either vmin or vmax are not specified, calculate an estimate by subsampling
                vmin_estimate, vmax_estimate = quick_min_max(d)

                # decide vmin, vmax passed to ImageGraphic constructor based on whether it's user specified or now
                if vmin_specified is None:
                    # user hasn't specified vmin, use estimated value
                    vmin = vmin_estimate
                else:
                    # user has provided a specific value, use that
                    vmin = vmin_specified

                if vmax_specified is None:
                    vmax = vmax_estimate
                else:
                    vmax = vmax_specified
            else:
                # both vmin and vmax are specified
                vmin, vmax = vmin_specified, vmax_specified

            ig = ImageGraphic(
                frame,
                name="image_widget_managed",
                vmin=vmin,
                vmax=vmax,
                **graphic_kwargs,
            )
            subplot.add_graphic(ig)

            if self._histogram_widget:
                hlut = HistogramLUTTool(data=d, images=ig, name="histogram_lut")

                subplot.docks["right"].add_graphic(hlut)
                subplot.docks["right"].size = 80
                subplot.docks["right"].auto_scale(maintain_aspect=False)
                subplot.docks["right"].controller.enabled = False

        # hard code the expected height so that the first render looks right in tests, docs etc.
        if len(self.slider_dims) == 0:
            ui_size = 57
        if len(self.slider_dims) == 1:
            ui_size = 106
        elif len(self.slider_dims) == 2:
            ui_size = 155

        self._image_widget_sliders = ImageWidgetSliders(
            figure=self.figure,
            size=ui_size,
            location="bottom",
            title="ImageWidget Controls",
            image_widget=self,
        )

        self.figure.add_gui(self._image_widget_sliders)

        self._current_index_changed_handlers = set()

        self._reentrant_block = False

        self._initialized = True

    @property
    def frame_apply(self) -> dict | None:
        return self._frame_apply

    @frame_apply.setter
    def frame_apply(self, frame_apply: dict[int, callable]):
        if frame_apply is None:
            frame_apply = dict()

        self._frame_apply = frame_apply
        # force update image graphic
        self.current_index = self.current_index

    @property
    def window_funcs(self) -> dict[str, _WindowFunctions]:
        """
        Get or set the window functions

        Returns
        -------
        Dict[str, _WindowFunctions]

        """
        return self._window_funcs

    @window_funcs.setter
    def window_funcs(self, callable_dict: dict[str, int]):
        if callable_dict is None:
            self._window_funcs = None
            # force frame to update
            self.current_index = self.current_index
            return

        elif isinstance(callable_dict, dict):
            if not set(callable_dict.keys()).issubset(ALLOWED_WINDOW_DIMS):
                raise ValueError(
                    f"The only allowed keys to window funcs are {list(ALLOWED_WINDOW_DIMS)} "
                    f"Your window func passed in these keys: {list(callable_dict.keys())}"
                )
            if not all(
                [
                    isinstance(_callable_dict, tuple)
                    for _callable_dict in callable_dict.values()
                ]
            ):
                raise TypeError(
                    "dict argument to `window_funcs` must be in the form of: "
                    "`{dimension: (func, window_size)}`. "
                    "See the docstring."
                )
            for v in callable_dict.values():
                if not callable(v[0]):
                    raise TypeError(
                        "dict argument to `window_funcs` must be in the form of: "
                        "`{dimension: (func, window_size)}`. "
                        "See the docstring."
                    )
                if not isinstance(v[1], int):
                    raise TypeError(
                        f"dict argument to `window_funcs` must be in the form of: "
                        "`{dimension: (func, window_size)}`. "
                        f"where window_size is integer. you passed in {v[1]} for window_size"
                    )

            if not isinstance(self._window_funcs, dict):
                self._window_funcs = dict()

            for k in list(callable_dict.keys()):
                self._window_funcs[k] = _WindowFunctions(self, *callable_dict[k])

        else:
            raise TypeError(
                f"`window_funcs` must be either Nonetype or dict."
                f"You have passed a {type(callable_dict)}. See the docstring."
            )

        # force frame to update
        self.current_index = self.current_index


    def _process_indices(
        self, array: np.ndarray, slice_indices: dict[str, int]
    ) -> np.ndarray:
        """
        Get the 2D array from the given slice indices. If not returning a 2D slice (such as due to window_funcs)
        then `frame_apply` must take this output and return a 2D array

        Parameters
        ----------
        array: np.ndarray
            array-like to get a 2D slice from

        slice_indices: Dict[str, int]
            dict in form of {dimension_index: current_index}
            For example if an array has shape [1000, 30, 512, 512] corresponding to [t, z, x, y]:
                To get the 100th timepoint and 3rd z-plane pass:
                    {"t": 100, "z": 3}

        Returns
        -------
        np.ndarray
            array-like, 2D slice

        """

        data_ix = None
        for i in range(len(self.data)):
            if self.data[i] is array:
                data_ix = i
                break

        numerical_dims = list()

        curr_ndim = self.data[data_ix].ndim

        indexer = [slice(None)] * curr_ndim

        curr_scrollable_format = SCROLLABLE_DIMS_ORDER[self.n_scrollable_dims[data_ix]]
        for dim in list(slice_indices.keys()):
            if dim not in curr_scrollable_format:
                continue
            numerical_dim = curr_scrollable_format.index(dim)
            indices_dim = slice_indices[dim]

            indices_dim = self._get_window_indices(data_ix, numerical_dim, indices_dim)

            indexer[numerical_dim] = indices_dim

            numerical_dims.append(numerical_dim)

        if self.window_funcs is not None:
            indexer = [slice(None)] * array.ndim
            window_axes = []

            # build indices before indexing
            for dim_str in curr_scrollable_format:
                if dim_str not in slice_indices:
                    continue

                numerical_dim = curr_scrollable_format.index(dim_str)
                idx = slice_indices[dim_str]
                idx = self._get_window_indices(data_ix, numerical_dim, idx)

                indexer[numerical_dim] = idx
                if not isinstance(idx, int):
                    window_axes.append((numerical_dim, dim_str))

            a = array[tuple(indexer)]

            for numerical_dim, dim_str in window_axes:
                func = self.window_funcs[dim_str].func
                effective_axis = numerical_dim - sum(ax < numerical_dim for ax, _ in window_axes if ax < numerical_dim)
                a = func(a, axis=effective_axis)

            return a
        else:
            return array[tuple(indexer)]

    def _get_window_indices(self, data_ix, dim, indices_dim):
        if self.window_funcs is None:
            return indices_dim

        else:
            ix = indices_dim

            dim_str = SCROLLABLE_DIMS_ORDER[self.n_scrollable_dims[data_ix]][dim]

            # if no window stuff specified for this dim
            if dim_str not in self.window_funcs.keys():
                return indices_dim

            # if window stuff is set to None for this dim
            # example: {"t": None}
            if self.window_funcs[dim_str] is None:
                return indices_dim

            window_size = self.window_funcs[dim_str].window_size

            if (window_size == 0) or (window_size is None):
                return indices_dim

            half_window = int((window_size - 1) / 2)  # half-window size
            # get the max bound for that dimension
            max_bound = self._dims_max_bounds[dim_str]
            indices_dim = range(
                max(0, ix - half_window), min(max_bound, ix + half_window)
            )
            return indices_dim

    def _process_frame_apply(self, array, data_ix) -> np.ndarray:
        if callable(self._frame_apply):
            return self._frame_apply(array)

        if data_ix not in self._frame_apply.keys():
            return array

        elif self._frame_apply[data_ix] is not None:
            return self._frame_apply[data_ix](array)

        return array

    def add_event_handler(self, handler: callable, event: str = "current_index"):
        """
        Register an event handler.

        Currently the only event that ImageWidget supports is "current_index". This event is
        emitted whenever the index of the ImageWidget changes.

        Parameters
        ----------
        handler: callable
            callback function, must take a dict as the only argument. This dict will be the `current_index`

        event: str, "current_index"
            the only supported event is "current_index"

        Example
        -------

        .. code-block:: py

            def my_handler(index):
                print(index)
                # example prints: {"t": 100} if data has only time dimension
                # "z" index will be another key if present in the data, ex: {"t": 100, "z": 5}

            # create an image widget
            iw = ImageWidget(...)

            # add event handler
            iw.add_event_handler(my_handler)

        """
        if event != "current_index":
            raise ValueError(
                "`current_index` is the only event supported by `ImageWidget`"
            )

        self._current_index_changed_handlers.add(handler)

    def remove_event_handler(self, handler: callable):
        """Remove a registered event handler"""
        self._current_index_changed_handlers.remove(handler)

    def clear_event_handlers(self):
        """Clear all registered event handlers"""
        self._current_index_changed_handlers.clear()

    def reset_vmin_vmax(self):
        """
        Reset the vmin and vmax w.r.t. the full data
        """
        for data, subplot in zip(self.data, self.figure):
            if "histogram_lut" not in subplot.docks["right"]:
                continue
            hlut = subplot.docks["right"]["histogram_lut"]
            hlut.set_data(data, reset_vmin_vmax=True)

    def reset_vmin_vmax_frame(self):
        """
        Resets the vmin vmax and HistogramLUT widgets w.r.t. the current data shown in the
        ImageGraphic instead of the data in the full data array. For example, if a post-processing
        function is used, the range of values in the ImageGraphic can be very different from the
        range of values in the full data array.

        TODO: We could think of applying the frame_apply funcs to a subsample of the entire array to get a better estimate of vmin vmax?
        """

        for subplot in self.figure:
            if "histogram_lut" not in subplot.docks["right"]:
                continue

            hlut = subplot.docks["right"]["histogram_lut"]
            # set the data using the current image graphic data
            hlut.set_data(subplot["image_widget_managed"].data.value)

    def set_data(
        self,
        new_data: np.ndarray | list[np.ndarray],
        reset_vmin_vmax: bool = True,
        reset_indices: bool = True,
    ):
        """
        Change data of widget. Note: sliders max currently update only for ``txy`` and ``tzxy`` data.

        Parameters
        ----------
        new_data: array-like or list of array-like
            The new data to display in the widget

        reset_vmin_vmax: bool, default ``True``
            reset the vmin vmax levels based on the new data

        reset_indices: bool, default ``True``
            reset the current index for all dimensions to 0

        """

        if reset_indices:
            for key in self.current_index:
                self.current_index[key] = 0

        # set slider max according to new data
        max_lengths = dict()
        for scroll_dim in self.slider_dims:
            max_lengths[scroll_dim] = np.inf

        if _is_arraylike(new_data):
            new_data = [new_data]

        if len(self._data) != len(new_data):
            raise ValueError(
                f"number of new data arrays {len(new_data)} must match"
                f" current number of data arrays {len(self._data)}"
            )
        # check all arrays
        for i, (new_array, current_array) in enumerate(zip(new_data, self._data)):
            if new_array.ndim != current_array.ndim:
                raise ValueError(
                    f"new data ndim {new_array.ndim} at index {i} "
                    f"does not equal current data ndim {current_array.ndim}"
                )

            # Computes the number of scrollable dims and also validates new_array
            new_scrollable_dims = self._get_n_scrollable_dims(new_array, self._rgb[i])

            if self.n_scrollable_dims[i] != new_scrollable_dims:
                raise ValueError(
                    f"number of dimensions of data arrays must match number of dimensions of "
                    f"existing data arrays"
                )

        # if checks pass, update with new data
        for i, (new_array, current_array, subplot) in enumerate(
            zip(new_data, self._data, self.figure)
        ):
            # check last two dims (x and y) to see if data shape is changing
            old_data_shape = self._data[i].shape[-self.n_img_dims[i] :]
            self._data[i] = new_array

            if old_data_shape != new_array.shape[-self.n_img_dims[i] :]:
                frame = self._process_indices(
                    new_array, slice_indices=self._current_index
                )
                frame = self._process_frame_apply(frame, i)

                # make new graphic first
                new_graphic = ImageGraphic(data=frame, name="image_widget_managed")

                if self._histogram_widget:
                    # set hlut tool to use new graphic
                    subplot.docks["right"]["histogram_lut"].images = new_graphic

                # delete old graphic after setting hlut tool to new graphic
                # this ensures gc
                subplot.delete_graphic(graphic=subplot["image_widget_managed"])
                subplot.insert_graphic(graphic=new_graphic)

            # Returns "", "t", or "tz"
            curr_scrollable_format = SCROLLABLE_DIMS_ORDER[self.n_scrollable_dims[i]]

            for scroll_dim in self.slider_dims:
                if scroll_dim in curr_scrollable_format:
                    new_length = new_array.shape[
                        curr_scrollable_format.index(scroll_dim)
                    ]
                    if max_lengths[scroll_dim] == np.inf:
                        max_lengths[scroll_dim] = new_length
                    elif max_lengths[scroll_dim] != new_length:
                        raise ValueError(
                            f"New arrays have differing values along dim {scroll_dim}"
                        )

                    self._dims_max_bounds[scroll_dim] = max_lengths[scroll_dim]

            # set histogram widget
            if self._histogram_widget:
                subplot.docks["right"]["histogram_lut"].set_data(
                    new_array, reset_vmin_vmax=reset_vmin_vmax
                )

        # force graphics to update
        self.current_index = self.current_index

    def show(self, **kwargs):
        """
        Show the widget.

        Parameters
        ----------

        kwargs: Any
            passed to `Figure.show()`

        Returns
        -------
        BaseRenderCanvas
            In Qt or GLFW, the canvas window containing the Figure will be shown.
            In jupyter, it will display the plot in the output cell or sidecar.

        """

        return self.figure.show(**kwargs)

    def close(self):
        """Close Widget"""
        self.figure.close()



from warnings import warn
from typing import Literal

import numpy as np
from numpy.typing import NDArray

from rendercanvas.base import log_exception

import pygfx


def to_gpu_supported_dtype(array):
    """
    convert input array to float32 numpy array
    """
    if isinstance(array, np.ndarray):
        if not array.dtype == np.float32:
            warn(f"casting {array.dtype} array to float32")
            return array.astype(np.float32)
        return array

    # try to make a numpy array from it, should not copy, tested with jax arrays
    return np.asarray(array).astype(np.float32)


class GraphicFeatureEvent(pygfx.Event):
    """
    **All event instances have the following attributes**

    +------------+-------------+-----------------------------------------------+
    | attribute  | type        | description                                   |
    +============+=============+===============================================+
    | type       | str         | "colors" - name of the event                  |
    +------------+-------------+-----------------------------------------------+
    | graphic    | Graphic     | graphic instance that the event is from       |
    +------------+-------------+-----------------------------------------------+
    | info       | dict        | event info dictionary                         |
    +------------+-------------+-----------------------------------------------+
    | target     | WorldObject | pygfx rendering engine object for the graphic |
    +------------+-------------+-----------------------------------------------+
    | time_stamp | float       | time when the event occurred, in ms           |
    +------------+-------------+-----------------------------------------------+

    """

    def __init__(self, type: str, info: dict):
        super().__init__(type=type)
        self.info = info


class GraphicFeature:
    def __init__(self, **kwargs):
        self._event_handlers = list()
        self._block_events = False

        # used by @block_reentrance decorator to block re-entrance into set_value functions
        self._reentrant_block: bool = False

    @property
    def value(self):
        """Graphic Feature value, must be implemented in subclass"""
        raise NotImplemented

    def set_value(self, graphic, value: float):
        """Graphic Feature value setter, must be implemented in subclass"""
        raise NotImplementedError

    def block_events(self, val: bool):
        """
        Block all events from this feature

        Parameters
        ----------
        val: bool
            ``True`` or ``False``

        """
        self._block_events = val

    def add_event_handler(self, handler: callable):
        """
        Add an event handler. All added event handlers are called when this feature changes.

        Used by `Graphic` classes to add to their event handlers, not meant for users. Users
        add handlers to Graphic instances only.

        The ``handler`` must accept a :class:`.FeatureEvent` as the first and only argument.

        Parameters
        ----------
        handler: callable
            a function to call when this feature changes

        """
        if not callable(handler):
            raise TypeError("event handler must be callable")

        if handler in self._event_handlers:
            warn(f"Event handler {handler} is already registered.")
            return

        self._event_handlers.append(handler)

    def remove_event_handler(self, handler: callable):
        """
        Remove a registered event ``handler``.

        Parameters
        ----------
        handler: callable
            event handler to remove

        """
        if handler not in self._event_handlers:
            raise KeyError(f"event handler {handler} not registered.")

        self._event_handlers.remove(handler)

    def clear_event_handlers(self):
        """Clear all event handlers"""
        self._event_handlers.clear()

    def _call_event_handlers(self, event_data: GraphicFeatureEvent):
        if self._block_events:
            return

        for func in self._event_handlers:
            with log_exception(
                f"Error during handling {self.__class__.__name__} event"
            ):
                func(event_data)


class BufferManager(GraphicFeature):
    """Smaller wrapper for pygfx.Buffer"""

    def __init__(
        self,
        data: NDArray | pygfx.Buffer,
        buffer_type: Literal["buffer", "texture", "texture-array"] = "buffer",
        isolated_buffer: bool = True,
        texture_dim: int = 2,
        **kwargs,
    ):
        super().__init__()
        if isolated_buffer and not isinstance(data, pygfx.Resource):
            # useful if data is read-only, example: memmaps
            bdata = np.zeros(data.shape, dtype=data.dtype)
            bdata[:] = data[:]
        else:
            # user's input array is used as the buffer
            bdata = data

        if isinstance(data, pygfx.Resource):
            # already a buffer, probably used for
            # managing another BufferManager, example: VertexCmap manages VertexColors
            self._buffer = data
        elif buffer_type == "buffer":
            self._buffer = pygfx.Buffer(bdata)
        elif buffer_type == "texture":
            # TODO: placeholder, not currently used since TextureArray is used specifically for Image graphics
            self._buffer = pygfx.Texture(bdata, dim=texture_dim)
        else:
            raise ValueError(
                "`data` must be a pygfx.Buffer instance or `buffer_type` must be one of: 'buffer' or 'texture'"
            )

        self._event_handlers: list[callable] = list()

        self._shared: int = 0

    @property
    def value(self) -> np.ndarray:
        """numpy array object representing the data managed by this buffer"""
        return self.buffer.data

    def set_value(self, graphic, value):
        """Sets values on entire array"""
        self[:] = value

    @property
    def buffer(self) -> pygfx.Buffer | pygfx.Texture:
        """managed buffer"""
        return self._buffer

    @property
    def shared(self) -> int:
        """Number of graphics that share this buffer"""
        return self._shared

    @property
    def __array_interface__(self):
        raise BufferError(
            f"Cannot use graphic feature buffer as an array, use <feature-name>.value instead.\n"
            f"Examples: line.data.value, line.colors.value, scatter.data.value, scatter.sizes.value"
        )

    def __getitem__(self, item):
        return self.buffer.data[item]

    def __setitem__(self, key, value):
        raise NotImplementedError

    def _parse_offset_size(
        self,
        key: int | slice | np.ndarray[int | bool] | list[bool | int],
        upper_bound: int,
    ):
        """
        parse offset and size for first, i.e. n_datapoints, dimension
        """
        if np.issubdtype(type(key), np.integer):
            # simplest case, just an int
            offset = key
            size = 1

        elif isinstance(key, slice):
            # TODO: off-by-one sometimes when step is used
            #  the offset can be one to the left or the size
            #  is one extra so it's not really an issue for now
            # parse slice
            start, stop, step = key.indices(upper_bound)

            # account for backwards indexing
            if (start > stop) and step < 0:
                offset = stop
            else:
                offset = start

            # slice.indices will give -1 if None is passed
            # which just means 0 here since buffers do not
            # use negative indexing
            offset = max(0, offset)

            # number of elements to upload
            # this is indexing so do not add 1
            size = abs(stop - start)

        elif isinstance(key, (np.ndarray, list)):
            if isinstance(key, list):
                # convert to array
                key = np.array(key)

            if not key.ndim == 1:
                raise TypeError(
                    f"can only use 1D arrays for fancy indexing, you have passed a data with: {key.ndim} dimensions"
                )

            if key.dtype == bool:
                # convert bool mask to integer indices
                key = np.nonzero(key)[0]

            if not np.issubdtype(key.dtype, np.integer):
                # fancy indexing doesn't make sense with non-integer types
                raise TypeError(
                    f"can only using integer or booleans arrays for fancy indexing, your array is of type: {key.dtype}"
                )

            if key.size < 1:
                # nothing to update
                return

            # convert any negative integer indices to positive indices
            key %= upper_bound

            # index of first element to upload
            offset = key.min()

            # size range to upload
            # add 1 because this is direct
            # passing of indices, not a start:stop
            size = np.ptp(key) + 1

        else:
            raise TypeError(
                f"invalid key for indexing buffer: {key}\n"
                f"valid ways to index buffers are using integers, slices, or fancy indexing with integers or bool"
            )

        return offset, size

    def _update_range(
        self,
        key: (
            int | slice | np.ndarray[int | bool] | list[bool | int] | tuple[slice, ...]
        ),
    ):
        """
        Uses key from slicing to determine the offset and
        size of the buffer to mark for upload to the GPU
        """
        upper_bound = self.value.shape[0]

        if isinstance(key, tuple):
            if any([k is Ellipsis for k in key]):
                # let's worry about ellipsis later
                raise TypeError("ellipses not supported for indexing buffers")
            # if multiple dims are sliced, we only need the key for
            # the first dimension corresponding to n_datapoints
            key: int | np.ndarray[int | bool] | slice = key[0]

        offset, size = self._parse_offset_size(key, upper_bound)
        self.buffer.update_range(offset=offset, size=size)

    def _emit_event(self, type: str, key, value):
        if len(self._event_handlers) < 1:
            return

        event_info = {
            "key": key,
            "value": value,
        }
        event = GraphicFeatureEvent(type, info=event_info)

        self._call_event_handlers(event)

    def __len__(self):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__} buffer data:\n{self.value.__repr__()}"


def block_reentrance(set_value):
    # decorator to block re-entrant set_value methods
    # useful when creating complex, circular, bidirectional event graphs
    def set_value_wrapper(self: GraphicFeature, graphic_or_key, value):
        """
        wraps GraphicFeature.set_value

        self: GraphicFeature instance

        graphic_or_key: graphic, or key if a BufferManager

        value: the value passed to set_value()
        """
        # set_value is already in the middle of an execution, block re-entrance
        if self._reentrant_block:
            return
        try:
            # block re-execution of set_value until it has *fully* finished executing
            self._reentrant_block = True
            set_value(self, graphic_or_key, value)
        except Exception as exc:
            # raise original exception
            raise exc  # set_value has raised. The line above and the lines 2+ steps below are probably more relevant!
        finally:
            # set_value has finished executing, now allow future executions
            self._reentrant_block = False

    return set_value_wrapper



from pathlib import Path

from ._version import __version__, version_info

# this must be the first import for auto-canvas detection
from .utils import loop  # noqa
from .graphics import *
from .graphics.features import GraphicFeatureEvent
from .graphics.selectors import *
from .graphics.utils import pause_events
from .legends import *
from .tools import *

from .layouts import IMGUI

if IMGUI:
    # default to imgui figure if imgui_bundle is installed
    from .layouts import ImguiFigure as Figure
else:
    from .layouts import Figure

from .widgets import ImageWidget
from .utils import config, enumerate_adapters, select_adapter, print_wgpu_report


if len(enumerate_adapters()) < 1:
    from warnings import warn

    warn(
        f"WGPU could not enumerate any adapters, fastplotlib will not work.\n"
        f"This is caused by one of the following:\n"
        f"1. You do not have a hardware GPU installed and you do not have "
        f"software rendering (ex. lavapipe) installed either\n"
        f"2. Your GPU drivers are not installed or something is wrong with your GPU driver installation, "
        f"re-installing the latest drivers from your hardware vendor (probably Nvidia or AMD) may help.\n"
        f"3. You are missing system libraries that are required for WGPU to access GPU(s), this is "
        f"common in cloud computing environments.\n"
        f"These two links can help you troubleshoot:\n"
        f"https://wgpu-py.readthedocs.io/en/stable/start.html#platform-requirements\n"
        f"https://fastplotlib.readthedocs.io/en/latest/user_guide/gpu.html\n",
        RuntimeWarning,
    )



import numpy as np
import pygfx

from ._base import GraphicFeature, GraphicFeatureEvent, block_reentrance


class Name(GraphicFeature):
    property_name = "name"
    event_info_spec = [
        {"dict key": "value", "type": "str", "description": "user provided name"},
    ]

    def __init__(self, value: str):
        """Graphic name"""

        self._value = value
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        if not isinstance(value, str):
            raise TypeError("`Graphic` name must be of type <str>")

        if graphic._plot_area is not None:
            graphic._plot_area._check_graphic_name_exists(value)

        self._value = value

        event = GraphicFeatureEvent(type="name", info={"value": value})
        self._call_event_handlers(event)


class Offset(GraphicFeature):
    property_name = "offset"
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray[float, float, float]",
            "description": "new offset (x, y, z)",
        },
    ]

    def __init__(self, value: np.ndarray | list | tuple):
        """Offset position of the graphic, [x, y, z]"""

        self._validate(value)
        # initialize zeros array
        self._value = np.zeros(3)

        # set values
        self._value[:] = value
        super().__init__()

    def _validate(self, value):
        if not len(value) == 3:
            raise ValueError("offset must be a list, tuple, or array of 3 float values")

    @property
    def value(self) -> np.ndarray:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: np.ndarray | list | tuple):
        self._validate(value)
        value = np.asarray(value)

        graphic.world_object.world.position = value

        # sometimes there are transforms so get the final position value like this
        value = graphic.world_object.world.position.copy()

        # set value of existing feature value array
        self._value[:] = value

        event = GraphicFeatureEvent(type="offset", info={"value": value})
        self._call_event_handlers(event)


class Rotation(GraphicFeature):
    property_name = "offset"
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray[float, float, float, float]",
            "description": "new rotation quaternion",
        },
    ]

    def __init__(self, value: np.ndarray | list | tuple):
        """Graphic rotation quaternion"""

        self._validate(value)
        # create zeros array
        self._value = np.zeros(4)

        self._value[:] = value
        super().__init__()

    def _validate(self, value):
        if not len(value) == 4:
            raise ValueError(
                "rotation quaternion must be a list, tuple, or array of 4 float values"
            )

    @property
    def value(self) -> np.ndarray:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: np.ndarray | list | tuple):
        self._validate(value)
        value = np.asarray(value)

        graphic.world_object.world.rotation = value

        # get the actual final quaternion value, pygfx adjusts to make sure || q ||_2 == 1
        # i.e. pygfx checks to make sure norm 1 and other transforms
        value = graphic.world_object.world.rotation.copy()

        # set value of existing feature value array
        self._value[:] = value

        event = GraphicFeatureEvent(type="rotation", info={"value": value})
        self._call_event_handlers(event)


class Alpha(GraphicFeature):
    """The alpha value (i.e. opacity) of a graphic."""

    property_name = "alpha"
    event_info_spec = [
        {"dict key": "value", "type": "float", "description": "new alpha value"},
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        wo = graphic.world_object
        if wo.material is not None:
            wo.material.opacity = value

        if "Image" in graphic.__class__.__name__:
            # Image and ImageVolume use tiling and share one material
            graphic._material.alpha = value

        self._value = value

        event = GraphicFeatureEvent(type="alpha", info={"value": value})
        self._call_event_handlers(event)


class AlphaMode(GraphicFeature):
    """The alpha-mode value of a graphic (i.e. how alpha is handled by the renderer)."""

    property_name = "alpha_mode"
    event_info_spec = [
        {"dict key": "value", "type": "str", "description": "new alpha mode"},
    ]

    def __init__(self, value: str):
        self._value = value
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        wo = graphic.world_object
        if wo.material is not None:
            wo.alpha_mode = value

        if "Image" in graphic.__class__.__name__:
            # Image and ImageVolume use tiling and share one material
            graphic._material.alpha_mode = value

        self._value = value

        event = GraphicFeatureEvent(type="alpha_mode", info={"value": value})
        self._call_event_handlers(event)


class Visible(GraphicFeature):
    """Access or change the visibility."""

    property_name = "visible"
    event_info_spec = [
        {"dict key": "value", "type": "bool", "description": "new visibility bool"},
    ]

    def __init__(self, value: bool):
        self._value = value
        super().__init__()

    @property
    def value(self) -> bool:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: bool):
        graphic.world_object.visible = value
        self._value = value

        event = GraphicFeatureEvent(type="visible", info={"value": value})
        self._call_event_handlers(event)


class Deleted(GraphicFeature):
    """
    Used when a graphic is deleted, triggers events that can be useful to indicate this graphic has been deleted
    """

    property_name = "deleted"
    event_info_spec = [
        {
            "dict key": "value",
            "type": "bool",
            "description": "True when graphic was deleted",
        },
    ]

    def __init__(self, value: bool):
        self._value = value
        super().__init__()

    @property
    def value(self) -> bool:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: bool):
        self._value = value
        event = GraphicFeatureEvent(type="deleted", info={"value": value})
        self._call_event_handlers(event)
